part of sliders;

/// Signature for formatting or changing the whole numeric or date label text.
typedef LabelFormatterCallback = String Function(

    /// actualValue will be either [DateTime] or [double] based on given [values].
    dynamic actualValue,

    /// If the actual value is [double], it is formatted by [numberFormat] and
    /// if the actual value is [DateTime], it is formatted by [dateFormat].
    String formattedText);

/// Signature for formatting or changing the whole tooltip label text.
typedef TooltipTextFormatterCallback = String Function(

    /// actualValue will be either [DateTime] or [double] based on given [values].
    dynamic actualValue,

    /// If the actual value is [double], it is formatted by [numberFormat] and
    /// if the actual value is [DateTime], it is formatted by [dateFormat].
    String formattedText);

/// Option to place the labels either between the major ticks or on the major ticks.
enum LabelPlacement {
  /// onTicks places the labels on the major ticks.
  onTicks,

  /// betweenTicks places the labels between the major ticks.
  betweenTicks
}

/// The type of date interval. It can be years to seconds.
enum DateIntervalType {
  /// Date interval is year.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 00) and [max] is DateTime(2005, 12, 31, 24)
  /// and [interval] is 1 and [dateIntervalType] is [years]
  /// then range slider will render labels for 2000, 2001, 2002, 2003, 2004, 2005 respectively.
  years,

  /// Date interval is month.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 00) and [max] is DateTime(2000, 12, 31, 24)
  /// and [interval] is 3 and [dateIntervalType] is [months] then range slider will render labels
  /// for [Jan 01, 2000], [Apr 01, 2000], [Jul 01, 2000], [Oct 01, 2000] and [Jan 01, 2001] respectively.
  months,

  /// Date interval is day.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 00) and [max] is DateTime(2000, 01, 25, 24)
  /// and [interval] is 5 and [dateIntervalType] is [days] then range slider will render labels
  /// for [Jan 01, 2000], [Jan 06, 2000], [Jan 11, 2000], [Jan 16, 2000], [Jan 21, 2001] and [Jan 26, 2001] respectively.
  days,

  /// Date interval is hour.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 09) and [max] is DateTime(2000, 01, 01, 17) and
  /// [interval] is 4 and [dateIntervalType] is [hours] then range slider will render labels for
  /// [Jan 01, 2000 09:00], [Jan 01, 2000 13:00], and [Jan 01, 2000 17:00] respectively.
  hours,

  /// Date interval is minute.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 09) and [max] is DateTime(2000, 01, 01, 10) and
  /// [interval] is 15 and [dateIntervalType] is [minutes] then range slider will render labels for
  /// [Jan 01, 2000 09:00], [Jan 01, 2000 09:15], [Jan 01, 2000 09:30], [Jan 01, 2000 09:45]and [Jan 01, 2000 10:00] respectively.
  minutes,

  /// Date interval is second.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 09, 00) and [max] is DateTime(2000, 01, 01, 09, 01) and
  /// [interval] is 20 and [dateIntervalType] is [seconds] then range slider will render labels for
  /// [Jan 01, 2000 09:00:00], [Jan 01, 2000 09:00:20], [Jan 01, 2000 09:00:40], and [Jan 01, 2000 09:01:00] respectively.
  seconds
}

/// Represents the [SfRangeSlider] or [SfRangeSelector] thumbs.
enum SfThumb {
  /// start represents the [SfRangeValues.start] thumb.
  start,

  /// end represents the [SfRangeValues.end] thumb.
  end
}

const Offset _tooltipTextPadding = Offset(15, 15);

const double _tooltipTriangleHeight = 7;

const double _tooltipTriangleWidth = 12;

const double _minPreferredTouchWidth = 20;

/// Represents the current selected values of [SfRangeSlider] and [SfRangeSelector].
class SfRangeValues {
  /// Represents the current selected values of [SfRangeSlider] and [SfRangeSelector].
  const SfRangeValues(this.start, this.end);

  /// Represents the [SfRangeValues.start] thumb.
  final dynamic start;

  /// Represents the [SfRangeValues.end] thumb.
  final dynamic end;

  SfRangeValues _copyWith({dynamic start, dynamic end}) {
    // HACK: In web, 0.00 and 0 are considered as identical.
    // So, we had considered both double and int.
    if (start != null &&
        (start.runtimeType == double || start.runtimeType == int) &&
        this.end.runtimeType == DateTime) {
      final double value = start;
      return SfRangeValues(
          DateTime.fromMillisecondsSinceEpoch(value.toInt()), end ?? this.end);
    } else if (end != null &&
        (end.runtimeType == double || end.runtimeType == int) &&
        this.start.runtimeType == DateTime) {
      final double value = end;
      return SfRangeValues(start ?? this.start,
          DateTime.fromMillisecondsSinceEpoch(value.toInt()));
    }

    return SfRangeValues(start ?? this.start, end ?? this.end);
  }
}
