part of sliders;

/// A Material Design range selector.
///
/// Used to select a range between [min] and [max]. It supports both numeric and date ranges. It accepts any type of framework widget or custom widget like [SfChart](https://pub.dev/documentation/syncfusion_flutter_charts/latest/charts/charts-library.html) as a child.
///
/// The range selector elements are:
///
/// * The "track", which is the rounded rectangle in which the thumbs are slides over.
/// * The "thumb", which is a shape that slides horizontally when the user drags it.
/// * The "active" side of the range selector is between the left and right thumbs.
/// * The "inactive" side of the range selector is between the [min] value and the left thumb,
/// and the right thumb and the [max] value. For RTL, the inactive side of the range selector is between
/// the [max] value and the left thumb, and the right thumb and the [min] value.
/// * The "divisors", which is a shape that renders on the track based on the
/// given [interval] value.
/// * The "ticks", which is a shape that rendered based on given [interval] value.
/// Basically, it is rendered below the track. It is also called “major ticks”.
/// * The "minor ticks", which is a shape that renders between two major ticks based on given [minorTicksPerInterval] value.
/// Basically, it is rendered below the track.
/// * The "labels", which is a text that rendered based on given [interval] value.
/// Basically, it is rendered below the track and the major ticks.
/// * The "child", which can be any type of framework widget or custom widget like [SfChart](https://pub.dev/documentation/syncfusion_flutter_charts/latest/charts/charts-library.html) as a child.
///
/// Either [controller] or [initialValues] need to be set to render [SfRangeSelector].
///
/// The range selector will be disabled if [enabled] is false or [min] is equal to [max].
///
/// The range selector widget maintains state internally.
/// So, the widget calls the [onChanged] callback with the new values
/// when the state of range selector changes.
///
/// Range selector can be customized using the [SfRangeSliderTheme](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderTheme-class.html) with the help of [SfRangeSelectorThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), or the [SfTheme](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfTheme-class.html) with the help of [SfThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfThemeData-class.html).
/// It is also possible to override the appearance using [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) which were set using the properties of the widget like [activeColor] and [inactiveColor].
///
/// ## Example
///
/// This snippet shows how to create a numeric [SfRangeSelector].
///
/// ```dart
/// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
///
/// SfRangeSelector(
///     min: 0.0,
///     max: 10.0,
///     initialValues: _initialValues,
///     interval: 1,
///     showTicks: true,
///     showLabels: true,
///     child: Container(
///         height: 200,
///         color: Colors.green[100],
///     ),
/// )
/// ```
///
/// This snippet shows how to create a date [SfRangeSelector].
///
/// ```dart
/// SfRangeValues _initialValues = SfRangeValues(DateTime(2002, 12, 31, 24), DateTime(2003, 12, 31, 24));
///
/// SfRangeSelector(
///     min: DateTime(2000, 01, 01, 00),
///     max: DateTime(2005, 12, 31, 24),
///     initialValues: _initialValues,
///     interval: 1,
///     dateFormat: DateFormat.y(),
///     dateIntervalType: DateIntervalType.years,
///     child: Container(
///         height: 200,
///         color: Colors.green[100],
///     ),
/// )
/// ```
///
/// See also:
///
/// * [SfTheme](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfTheme-class.html) and [SfThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfThemeData-class.html), for customizing the visual appearance of the range selector.
/// * [numberFormat] and [dateFormat], for formatting the numeric and date labels.
/// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the visual appearance of the range selector.
/// * [RangeController](https://pub.dev/documentation/syncfusion_flutter_core/latest/core/RangeController-class.html), for coordinating between [SfRangeSelector] and the widget which listens to it.
/// * [SfChart](https://pub.dev/documentation/syncfusion_flutter_charts/latest/charts/charts-library.html) and [RangeController](https://pub.dev/documentation/syncfusion_flutter_core/latest/core/RangeController-class.html), for selection and zooming.
class SfRangeSelector extends StatefulWidget {
  const SfRangeSelector(
      {Key key,
      this.min = 0.0,
      this.max = 1.0,
      this.initialValues,
      this.onChanged,
      this.controller,
      this.enabled = true,
      this.interval,
      this.stepSize,
      this.minorTicksPerInterval,
      this.showTicks = false,
      this.showLabels = false,
      this.showDivisors = false,
      this.showTooltip = false,
      this.lockRange = false,
      this.inactiveColor,
      this.activeColor,
      this.numberFormat,
      this.dateFormat,
      this.dateIntervalType,
      this.labelPlacement,
      this.labelFormatterCallback,
      this.tooltipTextFormatterCallback,
      this.trackShape,
      this.divisorShape,
      this.overlayShape,
      this.thumbShape,
      this.tickShape,
      this.minorTickShape,
      @required this.child})
      : assert(min != null),
        assert(max != null),
        assert(interval == null || interval > 0),
        assert(stepSize == null || stepSize > 0),
        assert(child != null),
        assert(controller != null || initialValues != null),
        super(key: key);

  /// The minimum value the user can select.
  ///
  /// Defaults to 0.0. Must be less than [max].
  final dynamic min;

  /// The maximum value the user can select.
  ///
  /// Defaults to 1.0. Must be greater than [min].
  final dynamic max;

  /// The values that initially selected in the range selector.
  ///
  /// The range selector's start and end thumbs are drawn corresponding to these values.
  ///
  /// For date values, the range selector doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType], and [dateFormat] for date values,
  /// if the labels, ticks, and divisors are needed.
  ///
  /// The range selector widget maintains state internally.
  /// So, the widget calls the [onChanged] callback with the new values
  /// when the state of range selector changes.
  ///
  /// This snippet shows how to create a numeric [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///     min: 0.0,
  ///     max: 10.0,
  ///     initialValues: _initialValues,
  ///     interval: 1,
  ///     showTicks: true,
  ///     showLabels: true,
  ///     child: Container(
  ///         height: 200,
  ///         color: Colors.green[100],
  ///     ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [RangeController](https://pub.dev/documentation/syncfusion_flutter_core/latest/core/RangeController-class.html), for coordinating between [SfRangeSelector] and the widget which listens to it.
  /// * [showTicks], to render major ticks at given interval.
  /// * [showLabels], to render labels at given interval.
  final SfRangeValues initialValues;

  /// Called when the user is selecting a new values for the selector by dragging.
  ///
  /// When the values of the range selector changes, the widget calls the
  /// [onChanged] callback with the new values.
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///     min: 1.0,
  ///     max: 10.0,
  ///     initialValues: _initialValues,
  ///     onChanged: (SfRangeValues values) {
  ///     }
  ///     child: Container(
  ///         height: 200,
  ///         color: Colors.green[100],
  ///     ),
  /// )
  /// ```
  final ValueChanged<SfRangeValues> onChanged;

  /// Coordinates between [SfRangeSelector] and the widget which listens to it.
  ///
  /// Built-in support for selection and zooming with [SfChart](https://pub.dev/documentation/syncfusion_flutter_charts/latest/charts/charts-library.html).
  ///
  /// Range controller contains [start] and [end] values.
  ///
  /// [start] - represents the currently selected start value of the range selector.
  /// The start thumb of the range selector was drawn corresponding to this value.
  ///
  /// [end] - represents the currently selected end value of the range selector.
  /// The end thumb of the range selector was drawn corresponding to this value.
  ///
  /// [start] and [end] can be either `double` or `DateTime`.
  ///
  /// ## Selection in [SfChart](https://pub.dev/documentation/syncfusion_flutter_charts/latest/charts/charts-library.html).
  ///
  /// ```dart
  /// class RangeSelectorPage extends StatefulWidget {
  ///   const RangeSelectorPage({ Key key }) : super(key: key);
  ///   @override
  ///   _RangeSelectorPageState createState() => _RangeSelectorPageState();
  /// }
  ///
  /// class _RangeSelectorPageState extends State<RangeSelectorPage> {
  ///  final double _min = 2.0;
  ///  final double _max = 10.0;
  ///  RangeController _rangeController;
  ///  SfRangeValues _values = SfRangeValues(4.0, 6.0);
  ///
  /// @override
  /// void initState() {
  ///   super.initState();
  ///     _rangeController = RangeController(
  ///     start: _values.start,
  ///     end: _values.end);
  /// }
  ///
  /// @override
  /// void dispose() {
  ///   _rangeController.dispose();
  ///   super.dispose();
  /// }
  ///
  /// final List<Data> chartData = <Data>[
  ///    Data(x:2.0, y: 2.2),
  ///    Data(x:3.0, y: 3.4),
  ///    Data(x:4.0, y: 2.8),
  ///    Data(x:5.0, y: 1.6),
  ///    Data(x:6.0, y: 2.3),
  ///    Data(x:7.0, y: 2.5),
  ///    Data(x:8.0, y: 2.9),
  ///    Data(x:9.0, y: 3.8),
  ///    Data(x:10.0, y: 3.7),
  /// ];
  ///
  /// @override
  /// Widget build(BuildContext context) {
  ///  return MaterialApp(
  ///      home: Scaffold(
  ///          body: Center(
  ///              child: SfRangeSelector(
  ///                    min: _min,
  ///                    max: _max,
  ///                    interval: 1,
  ///                    showTicks: true,
  ///                    showLabels: true,
  ///                   controller: _rangeController,
  ///                    child: Container(
  ///                    height: 130,
  ///                   child: SfCartesianChart(
  ///                        margin: const EdgeInsets.all(0),
  ///                        primaryXAxis: NumericAxis(minimum: _min,
  ///                            maximum: _max,
  ///                            isVisible: false),
  ///                        primaryYAxis: NumericAxis(isVisible: false),
  ///                        plotAreaBorderWidth: 0,
  ///                        series: <SplineAreaSeries<Data, double>>[
  ///                            SplineAreaSeries<Data, double>(
  ///                                selectionSettings: SelectionSettings(
  ///                                    enable: true,
  ///                                    selectionController: _rangeController),
  ///                               color: Color.fromARGB(255, 126, 184, 253),
  ///                                dataSource: chartData,
  ///                                    xValueMapper: (Data sales, _) => sales.x,
  ///                                    yValueMapper: (Data sales, _) => sales.y)
  ///                             ],
  ///                         ),
  ///                     ),
  ///                 ),
  ///             )
  ///         )
  ///     );
  ///   }
  /// }
  /// ```
  ///
  /// ## Zooming in [SfChart](https://pub.dev/documentation/syncfusion_flutter_charts/latest/charts/charts-library.html).
  ///
  /// ```dart
  /// class RangeZoomingPage extends StatefulWidget {
  ///   const RangeZoomingPage({ Key key }) : super(key: key);
  ///   @override
  ///   _RangeZoomingPageState createState() => _RangeZoomingPageState();
  /// }
  ///
  /// class _RangeZoomingPageState extends State<RangeZoomingPage> {
  ///  final double _min = 2.0;
  ///  final double _max = 10.0;
  ///  SfRangeValues _values = SfRangeValues(4.0, 6.0);
  ///  RangeController _rangeController;
  ///
  /// @override
  /// void initState() {
  ///   super.initState();
  ///     _rangeController = RangeController(
  ///     start: _values.start,
  ///     end: _values.end);
  /// }
  ///
  /// @override
  /// void dispose() {
  ///   _rangeController.dispose();
  ///   super.dispose();
  /// }
  ///
  /// final List<Data> chartData = <Data>[
  ///    Data(x:2.0, y: 2.2),
  ///    Data(x:3.0, y: 3.4),
  ///    Data(x:4.0, y: 2.8),
  ///    Data(x:5.0, y: 1.6),
  ///    Data(x:6.0, y: 2.3),
  ///    Data(x:7.0, y: 2.5),
  ///    Data(x:8.0, y: 2.9),
  ///    Data(x:9.0, y: 3.8),
  ///    Data(x:10.0, y: 3.7),
  /// ];
  ///
  /// @override
  /// Widget build(BuildContext context) {
  ///  return MaterialApp(
  ///      home: Scaffold(
  ///          body: Center(
  ///              child: SfRangeSelector(
  ///                    min: _min,
  ///                    max: _max,
  ///                    interval: 1,
  ///                    showTicks: true,
  ///                    showLabels: true,
  ///                   controller: _rangeController,
  ///                    child: Container(
  ///                    height: 130,
  ///                   child: SfCartesianChart(
  ///                        margin: const EdgeInsets.all(0),
  ///                        primaryXAxis: NumericAxis(minimum: _min,
  ///                            maximum: _max,
  ///                            isVisible: false,
  ///                            rangeController: _rangeController),
  ///                        primaryYAxis: NumericAxis(isVisible: false),
  ///                        plotAreaBorderWidth: 0,
  ///                        series: <SplineAreaSeries<Data, double>>[
  ///                            SplineAreaSeries<Data, double>(
  ///                               color: Color.fromARGB(255, 126, 184, 253),
  ///                                dataSource: chartData,
  ///                                    xValueMapper: (Data sales, _) => sales.x,
  ///                                    yValueMapper: (Data sales, _) => sales.y)
  ///                             ],
  ///                         ),
  ///                     ),
  ///                 ),
  ///             )
  ///         )
  ///     );
  ///   }
  /// }
  /// ```
  final RangeController controller;

  /// Controls the range selector’s state whether it is disabled or enabled.
  ///
  /// Defaults to `true`.
  ///
  /// This snippet shows how to create disable state [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///     min: 0.0,
  ///     max: 10.0,
  ///     initialValues: _initialValues,
  ///     enabled: false,
  ///     child: Container(
  ///         height: 200,
  ///         color: Colors.green[100],
  ///     ),
  /// )
  /// ```
  final bool enabled;

  /// Splits the range selector into given interval.
  /// It is mandatory if labels, major ticks and divisors are needed.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range selector will render the labels, major ticks,
  /// and divisors at 0.0, 2.0, 4.0 and so on.
  ///
  /// For date values, the range selector doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType],
  /// and [dateFormat] for date values, if labels, ticks, and divisors are needed.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 00) and [max] is DateTime(2005, 12, 31, 24),
  /// [interval] is 1.0, [dateFormat] is DateFormat.y(), and [dateIntervalType] is DateIntervalType.years,
  /// then the range selector will render the labels, major ticks, and divisors at 2000, 2001, 2002 and so on.
  ///
  /// Defaults to null. Must be greater than 0.
  ///
  /// This snippet shows how to set numeric interval in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 2,
  ///   showTicks: true,
  ///   showLabels: true,
  ///    child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    )
  /// )
  /// ```
  ///
  /// This snippet shows how to set date interval in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSelector(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///    child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [showDivisors], to render divisors at given interval.
  /// * [showTicks], to render major ticks at given interval.
  /// * [showLabels], to render labels at given interval.
  final double interval;

  /// Option to select discrete values.
  ///
  /// [stepSize] doesn’t work for [DateTime] range selector and
  /// [lockRange] doesn't works with [stepSize].
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [stepSize] is 2.0,
  /// the range selector will move the thumbs at 0.0, 2.0, 4.0 and so on.
  ///
  /// Defaults to `null`. Must be greater than 0.
  ///
  /// This snippet shows how to create discrete [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   stepSize: 2,
  ///   child: Container(
  ///     height: 200,
  ///     color: Colors.green[100],
  ///   )
  /// )
  /// ```
  final double stepSize;

  /// Number of smaller ticks between two major ticks.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range selector will render the major ticks at 0.0, 2.0, 4.0 and so on.
  /// If minorTicksPerInterval is 1, then smaller ticks will be rendered on 1.0, 3.0 and so on.
  ///
  /// Defaults to `null`. Must be greater than 0.
  ///
  /// This snippet shows how to show minor ticks in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 2,
  ///   showTicks: true,
  ///   minorTicksPerInterval: 1,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [showTicks] to render major ticks at given interval.
  /// * [minorTickShape] and [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) for customizing
  /// the minor tick’s visual appearance.
  final int minorTicksPerInterval;

  /// Option to render the major ticks on the track.
  ///
  /// It is a shape which is used to represent the major interval points of the track.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range selector will render the major ticks at 0.0, 2.0, 4.0 and so on.
  ///
  /// Defaults to `false`.
  ///
  /// This snippet shows how to show major ticks in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 2,
  ///   showTicks: true,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [tickShape] and [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) for customizing the
  /// major tick’s visual appearance.
  final bool showTicks;

  /// Option to render the labels on given interval.
  ///
  /// Defaults to `false`.
  ///
  /// This snippet shows how to show labels in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showLabels: true,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [numberFormat] and [dateFormat] for formatting the numeric and date labels.
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) to customize the appearance of the labels.
  final bool showLabels;

  /// Option to render the divisors on the track.
  ///
  /// It is a shape which is used to represent the major interval points of the track.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range selector will render the divisors at 0.0, 2.0, 4.0 and so on.
  ///
  /// Defaults to `false`.
  ///
  /// This snippet shows how to show divisors in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 2,
  ///   showDivisors: true,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///   ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [showTicks], to render major ticks at given interval.
  /// * [showLabels], to render labels at given interval.
  /// * [divisorShape] and [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) for customizing
  /// the divisor’s visual appearance.
  final bool showDivisors;

  /// Option to enable tooltips for both the thumbs.
  ///
  /// Used to clearly indicate the current selection of the ranges during interaction.
  ///
  /// By default, tooltip text is formatted with either [numberFormat] or [dateFormat].
  ///
  /// This snippet shows how to show tooltip in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   showTooltip: true,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [tooltipTextFormatterCallback], for changing the default tooltip text.
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the appearance of the tooltip text.
  final bool showTooltip;

  /// Locks the range between start and end thumbs.
  ///
  /// When lock range is enabled, you can move both the thumbs
  /// at the same time by dragging in the area between start and end thumbs.
  ///
  /// [stepSize] doesn't works with [lockRange].
  ///
  /// Defaults to`false`.
  ///
  /// This snippet shows how to show tooltip in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   lockRange: true,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  final bool lockRange;

  /// Color applied to the inactive track and active divisors.
  ///
  /// The inactive side of the range selector is between the [min] value and
  /// the left thumb, and the right thumb and the [max] value.
  ///
  /// For RTL, the inactive side of the range selector is between the [max] value and
  /// the left thumb, and the right thumb and the [min] value.
  ///
  /// This snippet shows how to set inactive color in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   inactiveColor: Colors.red,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the individual
  /// inactive range selector element’s visual.
  final Color inactiveColor;

  /// Color applied to the active track, thumb, overlay, and inactive divisors.
  ///
  /// The active side of the range selector is between the start and end thumbs.
  ///
  /// This snippet shows how to set active color in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   activeColor: Colors.red,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the individual active range selector element’s visual.
  final Color activeColor;

  /// Formats the numeric labels.
  ///
  /// Defaults to `null`.
  ///
  /// This snippet shows how to set number format in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSelector(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   initialValues: _initialValues,
  ///   interval: 2,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   numberFormat: NumberFormat(‘\$’),
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  /// See also:
  ///
  /// * [labelFormatterCallback], for formatting the numeric and date labels.
  final NumberFormat numberFormat;

  /// Formats the date labels. It is mandatory for date [SfRangeSelector].
  ///
  /// For date values, the range selector doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType],
  /// and [dateFormat] for date values, if labels, ticks, and divisors are needed.
  ///
  /// Defaults to `null`.
  ///
  /// This snippet shows how to set date format in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSelector(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   showLabels: true,
  ///   showTicks: true,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [interval], for setting the interval.
  /// * [numberFormat], for formatting the numeric labels.
  /// * [labelFormatterCallback], for formatting the numeric and date label.
  /// * [dateIntervalType], for changing the interval type.
  final DateFormat dateFormat;

  /// The type of date interval. It is mandatory for date [SfRangeSelector].
  ///
  /// It can be years to seconds.
  ///
  /// For date values, the range selector doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType],
  /// and [dateFormat] for date values, if labels, ticks, and divisors are needed.
  ///
  /// Defaults to `null`.
  ///
  /// This snippet shows how to set date interval type in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSelector(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  final DateIntervalType dateIntervalType;

  /// Option to place the labels either between the major ticks or on the major ticks.
  ///
  /// Defaults to [LabelPlacement.onTicks].
  ///
  /// This snippet shows how to set label placement in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSelector(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   initialValues: _initialValues,
  ///   interval: 1,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   labelPlacement: LabelPlacement.betweenTicks,
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///    ),
  /// )
  /// ```
  final LabelPlacement labelPlacement;

  /// Signature for formatting or changing the whole numeric or date label text.
  ///
  /// * The actual value without formatting is given by `actualValue`. It is either [DateTime] or [double] based on given [initialValues] or controller start and end values.
  /// * The formatted value based on the numeric or date format is given by `formattedText`.
  ///
  /// This snippet shows how to set label format in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(100.0, 10000.0);
  ///
  /// SfRangeSelector(
  ///   min: 100.0,
  ///   max: 10000.0,
  ///   initialValues: _initialValues,
  ///   interval: 9000.0,
  ///   showLabels: true,
  ///   showTicks: true,
  ///   labelFormatterCallback: (dynamic actualValue, String formattedText) {
  ///     return actualValue == 10000 ? '\$ $ formattedText +' : '\$ $ formattedText';
  ///   },
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///   ),
  /// )
  /// ```
  final LabelFormatterCallback labelFormatterCallback;

  /// Signature for formatting or changing the whole tooltip label text.
  ///
  /// * The actual value without formatting is given by `actualValue`. It is either [DateTime] or [double] based on given [initialValues] or controller start and end values.
  /// * The formatted value based on the numeric or date format is given by `formattedText`.
  ///
  /// This snippet shows how to set tooltip format in [SfRangeSelector].
  ///
  /// ```dart
  /// SfRangeValues _initialValues = SfRangeValues(DateTime(2010, 01, 01, 13, 00, 00), DateTime(2010, 01, 01, 17, 00, 00));
  ///
  /// SfRangeSelector(
  ///   min: DateTime(2010, 01, 01, 9, 00, 00),
  ///   max: DateTime(2010, 01, 01, 21, 05, 00),
  ///   initialValues: _initialValues,
  ///   interval: 4,
  ///   showLabels: true,
  ///   showTooltip: true,
  ///   dateFormat: DateFormat('h a'),
  ///   dateIntervalType: DateIntervalType.hours,
  ///   tooltipTextFormatterCallback: (dynamic actualValue, String formattedText) {
  ///     return DateFormat('h:mm a').format(actualValue);
  ///   },
  ///   child: Container(
  ///       height: 200,
  ///       color: Colors.green[100],
  ///   ),
  /// )
  /// ```
  final TooltipTextFormatterCallback tooltipTextFormatterCallback;

  /// Base class for [SfRangeSelector] divisors shapes.
  final SfDivisorShape divisorShape;

  /// Base class for [SfRangeSelector] track shapes.
  final SfTrackShape trackShape;

  /// Base class for [SfRangeSelector] major tick shapes.
  final SfTickShape tickShape;

  /// Base class for [SfRangeSelector] minor tick shapes.
  final SfMinorTickShape minorTickShape;

  /// Base class for [SfRangeSelector] overlay shapes.
  final SfOverlayShape overlayShape;

  ///  Base class for [SfRangeSelector] thumb shapes.
  final SfThumbShape thumbShape;

  /// The content of range selector.
  ///
  /// If it is not null, [SfRangeSelectorThemeData]'s [activeRegionColor]
  /// and [inactiveRegionColor] applied on the child.
  ///
  /// The active side of the range selector is between the left and right thumb.
  /// The inactive side of the range selector is between the
  /// [min] value and left thumb as well as right thumb and [max] value.
  ///
  /// For RTL, The active side of the range selector is between the left and right thumb.
  /// The inactive side of the range selector is between the
  /// [min] value and right thumb as well as left thumb and [max] value.
  final Widget child;

  @override
  _SfRangeSelectorState createState() => _SfRangeSelectorState();
}

class _SfRangeSelectorState extends State<SfRangeSelector>
    with TickerProviderStateMixin {
  AnimationController overlayStartController;
  AnimationController overlayEndController;
  AnimationController stateController;
  final Duration duration = const Duration(milliseconds: 100);

  String _getFormattedLabelText(dynamic actualText, String formattedText) {
    return formattedText;
  }

  String _getFormattedTooltipText(dynamic actualText, String formattedText) {
    return formattedText;
  }

  @override
  void dispose() {
    overlayStartController?.dispose();
    overlayEndController?.dispose();
    stateController?.dispose();

    super.dispose();
  }

  @override
  void initState() {
    super.initState();

    if (widget.controller != null) {
      assert(widget.controller.start != null);
      assert(widget.controller.end != null);
    }

    overlayStartController =
        AnimationController(vsync: this, duration: duration);
    overlayEndController = AnimationController(vsync: this, duration: duration);
    stateController = AnimationController(vsync: this, duration: duration);
    stateController.value = widget.enabled ? 1.0 : 0.0;
  }

  @override
  Widget build(BuildContext context) {
    SyncfusionLicense.validateLicense(context);
    final ThemeData themeData = Theme.of(context);
    SfRangeSliderThemeData sliderThemeData = SfRangeSelectorTheme.of(context);
    sliderThemeData = sliderThemeData.copyWith(
        trackHeight: sliderThemeData.trackHeight,
        tickSize: sliderThemeData.tickSize,
        minorTickSize: sliderThemeData.minorTickSize,
        tickOffset: sliderThemeData.tickOffset,
        labelOffset: sliderThemeData.labelOffset ??
            (widget.showTicks
                ? const Offset(0.0, 5.0)
                : const Offset(0.0, 13.0)),
        inactiveLabelStyle: sliderThemeData.inactiveLabelStyle ??
            themeData.textTheme.body2.copyWith(
                color: widget.enabled
                    ? themeData.textTheme.body2.color.withOpacity(0.87)
                    : themeData.colorScheme.onSurface.withOpacity(0.32)),
        activeLabelStyle: sliderThemeData.activeLabelStyle ??
            themeData.textTheme.body2.copyWith(
                color: widget.enabled
                    ? themeData.textTheme.body2.color.withOpacity(0.87)
                    : themeData.colorScheme.onSurface.withOpacity(0.32)),
        tooltipTextStyle: sliderThemeData.tooltipTextStyle ??
            themeData.textTheme.body2.copyWith(
                color: sliderThemeData.brightness == Brightness.light
                    ? const Color.fromRGBO(255, 255, 255, 1)
                    : const Color.fromRGBO(0, 0, 0, 1)),
        inactiveTrackColor: widget.inactiveColor ??
            sliderThemeData.inactiveTrackColor ??
            themeData.primaryColor.withOpacity(0.24),
        activeTrackColor: widget.activeColor ??
            sliderThemeData.activeTrackColor ??
            themeData.primaryColor,
        thumbColor: widget.activeColor ??
            sliderThemeData.thumbColor ??
            themeData.primaryColor,
        activeTickColor: sliderThemeData.activeTickColor,
        inactiveTickColor: sliderThemeData.inactiveTickColor,
        disabledActiveTickColor: sliderThemeData.disabledActiveTickColor,
        disabledInactiveTickColor: sliderThemeData.disabledInactiveTickColor,
        activeMinorTickColor: sliderThemeData.activeMinorTickColor,
        inactiveMinorTickColor: sliderThemeData.inactiveMinorTickColor,
        disabledActiveMinorTickColor:
            sliderThemeData.disabledActiveMinorTickColor,
        disabledInactiveMinorTickColor:
            sliderThemeData.disabledInactiveMinorTickColor,
        overlayColor:
            widget.activeColor?.withOpacity(0.12) ?? sliderThemeData.overlayColor ?? themeData.primaryColor.withOpacity(0.12),
        inactiveDivisorColor: widget.activeColor ?? sliderThemeData.inactiveDivisorColor ?? themeData.colorScheme.primary.withOpacity(0.54),
        activeDivisorColor: widget.inactiveColor ?? sliderThemeData.activeDivisorColor ?? themeData.colorScheme.onPrimary.withOpacity(0.54),
        disabledInactiveDivisorColor: sliderThemeData.disabledInactiveDivisorColor ?? themeData.colorScheme.onSurface.withOpacity(0.12),
        disabledActiveDivisorColor: sliderThemeData.disabledActiveDivisorColor ?? themeData.colorScheme.onPrimary.withOpacity(0.12),
        disabledActiveTrackColor: sliderThemeData.disabledActiveTrackColor ?? themeData.colorScheme.onSurface.withOpacity(0.32),
        disabledInactiveTrackColor: sliderThemeData.disabledInactiveTrackColor ?? themeData.colorScheme.onSurface.withOpacity(0.12),
        disabledThumbColor: sliderThemeData.disabledThumbColor,
        activeRegionColor: sliderThemeData.activeRegionColor,
        inactiveRegionColor: widget.inactiveColor ?? sliderThemeData.inactiveRegionColor,
        tooltipBackgroundColor: sliderThemeData.tooltipBackgroundColor,
        trackCornerRadius: sliderThemeData.trackCornerRadius,
        overlayRadius: sliderThemeData.overlayRadius,
        thumbRadius: sliderThemeData.thumbRadius);

    return _RangeSelectorRenderObjectWidget(
      key: widget.key,
      min: widget.min,
      max: widget.max,
      values: widget.initialValues,
      enabled: widget.enabled,
      interval: widget.interval,
      stepSize: widget.stepSize,
      minorTicksPerInterval: widget.minorTicksPerInterval ?? 0,
      showTicks: widget.showTicks,
      showLabels: widget.showLabels,
      showDivisors: widget.showDivisors,
      showTooltip: widget.showTooltip,
      lockRange: widget.lockRange,
      inactiveColor:
          widget.inactiveColor ?? themeData.primaryColor.withOpacity(0.24),
      activeColor: widget.activeColor ?? themeData.primaryColor,
      labelPlacement: widget.labelPlacement ?? LabelPlacement.onTicks,
      numberFormat: widget.numberFormat ?? NumberFormat.decimalPattern(),
      dateFormat: widget.dateFormat,
      dateIntervalType: widget.dateIntervalType,
      labelFormatterCallback:
          widget.labelFormatterCallback ?? _getFormattedLabelText,
      tooltipTextFormatterCallback:
          widget.tooltipTextFormatterCallback ?? _getFormattedTooltipText,
      trackShape: widget.trackShape ?? _SfTrackShape(),
      divisorShape: widget.divisorShape ?? _SfDivisorShape(),
      overlayShape: widget.overlayShape ?? _SfOverlayShape(),
      thumbShape: widget.thumbShape ?? _SfThumbShape(),
      tickShape: widget.tickShape ?? _SfTickShape(),
      minorTickShape: widget.minorTickShape ?? _SfMinorTickShape(),
      child: widget.child,
      sliderThemeData: sliderThemeData,
      state: this,
    );
  }
}

class _RangeSelectorRenderObjectWidget extends SingleChildRenderObjectWidget {
  const _RangeSelectorRenderObjectWidget(
      {Key key,
      this.min,
      this.max,
      this.values,
      this.interval,
      this.stepSize,
      this.enabled,
      this.minorTicksPerInterval,
      this.showTicks,
      this.showLabels,
      this.showDivisors,
      this.showTooltip,
      this.lockRange,
      this.inactiveColor,
      this.activeColor,
      this.labelPlacement,
      this.numberFormat,
      this.dateFormat,
      this.dateIntervalType,
      this.labelFormatterCallback,
      this.tooltipTextFormatterCallback,
      this.trackShape,
      this.divisorShape,
      this.overlayShape,
      this.thumbShape,
      this.tickShape,
      this.minorTickShape,
      this.child,
      this.sliderThemeData,
      this.state})
      : super(key: key, child: child);

  final dynamic min;
  final dynamic max;
  final SfRangeValues values;
  final bool enabled;
  final double interval;
  final double stepSize;
  final int minorTicksPerInterval;

  final bool showTicks;
  final bool showLabels;
  final bool showDivisors;
  final bool showTooltip;
  final bool lockRange;

  final Color inactiveColor;
  final Color activeColor;

  final LabelPlacement labelPlacement;
  final NumberFormat numberFormat;
  final DateFormat dateFormat;
  final DateIntervalType dateIntervalType;
  final SfRangeSliderThemeData sliderThemeData;
  final LabelFormatterCallback labelFormatterCallback;
  final TooltipTextFormatterCallback tooltipTextFormatterCallback;
  final SfTrackShape trackShape;
  final SfDivisorShape divisorShape;
  final SfOverlayShape overlayShape;
  final SfThumbShape thumbShape;
  final SfTickShape tickShape;
  final SfMinorTickShape minorTickShape;
  final Widget child;
  final _SfRangeSelectorState state;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return _RenderRangeSelector(
      min: min,
      max: max,
      values: values,
      enabled: enabled,
      minorTicksPerInterval: minorTicksPerInterval,
      interval: interval,
      stepSize: stepSize,
      showTicks: showTicks,
      showLabels: showLabels,
      showDivisors: showDivisors,
      showTooltip: showTooltip,
      lockRange: lockRange,
      inactiveColor: inactiveColor,
      activeColor: activeColor,
      labelPlacement: labelPlacement,
      numberFormat: numberFormat,
      dateFormat: dateFormat,
      dateIntervalType: dateIntervalType,
      labelFormatterCallback: labelFormatterCallback,
      tooltipTextFormatterCallback: tooltipTextFormatterCallback,
      trackShape: trackShape,
      divisorShape: divisorShape,
      overlayShape: overlayShape,
      thumbShape: thumbShape,
      tickShape: tickShape,
      minorTickShape: minorTickShape,
      sliderThemeData: sliderThemeData,
      textDirection: Directionality.of(context),
      mediaQueryData: MediaQuery.of(context),
      state: state,
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, _RenderRangeSelector renderObject) {
    renderObject
      ..min = min
      ..max = max
      ..isEnabled = enabled
      ..interval = interval
      ..stepSize = stepSize
      ..minorTicksPerInterval = minorTicksPerInterval
      ..showTicks = showTicks
      ..showLabels = showLabels
      ..showDivisors = showDivisors
      ..showTooltip = showTooltip
      ..lockRange = lockRange
      ..inactiveColor = inactiveColor
      ..activeColor = activeColor
      ..labelPlacement = labelPlacement
      ..numberFormat = numberFormat
      ..dateFormat = dateFormat
      ..dateIntervalType = dateIntervalType
      ..labelFormatterCallback = labelFormatterCallback
      ..tooltipTextFormatterCallback = tooltipTextFormatterCallback
      ..trackShape = trackShape
      ..divisorShape = divisorShape
      ..overlayShape = overlayShape
      ..thumbShape = thumbShape
      ..tickShape = tickShape
      ..minorTickShape = minorTickShape
      ..sliderThemeData = sliderThemeData
      ..textDirection = Directionality.of(context)
      ..mediaQueryData = MediaQuery.of(context);
  }
}

class _RenderRangeSelector extends RenderProxyBox
    with RelayoutWhenSystemFontsChangeMixin, _SliderMixin {
  _RenderRangeSelector({
    dynamic min,
    dynamic max,
    SfRangeValues values,
    bool enabled,
    double interval,
    double stepSize,
    int minorTicksPerInterval,
    bool showTicks,
    bool showLabels,
    bool showDivisors,
    bool showTooltip,
    bool lockRange,
    Color inactiveColor,
    Color activeColor,
    LabelPlacement labelPlacement,
    NumberFormat numberFormat,
    DateFormat dateFormat,
    DateIntervalType dateIntervalType,
    LabelFormatterCallback labelFormatterCallback,
    TooltipTextFormatterCallback tooltipTextFormatterCallback,
    SfTrackShape trackShape,
    SfDivisorShape divisorShape,
    SfOverlayShape overlayShape,
    SfThumbShape thumbShape,
    SfTickShape tickShape,
    SfMinorTickShape minorTickShape,
    SfRangeSliderThemeData sliderThemeData,
    TextDirection textDirection,
    MediaQueryData mediaQueryData,
    @required _SfRangeSelectorState state,
  })  : _min = min,
        _max = max,
        _values = values,
        _isEnabled = enabled,
        _interval = interval,
        _stepSize = stepSize,
        _minorTicksPerInterval = minorTicksPerInterval,
        _showTicks = showTicks,
        _showLabels = showLabels,
        _showDivisors = showDivisors,
        _showTooltip = showTooltip,
        _lockRange = lockRange,
        _inactiveColor = inactiveColor,
        _activeColor = activeColor,
        _labelPlacement = labelPlacement,
        _numberFormat = numberFormat,
        _dateIntervalType = dateIntervalType,
        _dateFormat = dateFormat,
        _labelFormatterCallback = labelFormatterCallback,
        _tooltipTextFormatterCallback = tooltipTextFormatterCallback,
        _trackShape = trackShape,
        _divisorShape = divisorShape,
        _overlayShape = overlayShape,
        _thumbShape = thumbShape,
        _tickShape = tickShape,
        _minorTickShape = minorTickShape,
        _sliderThemeData = sliderThemeData,
        _textDirection = textDirection,
        _mediaQueryData = mediaQueryData,
        _state = state {
    if (_state.widget.controller != null) {
      assert(_state.widget.controller.start != null);
      assert(_state.widget.controller.end != null);
      _values = SfRangeValues(
          _state.widget.controller.start, _state.widget.controller.end);
    }

    final GestureArenaTeam team = GestureArenaTeam();

    _dragGestureRecognizer = HorizontalDragGestureRecognizer()
      ..team = team
      ..onStart = _onDragStart
      ..onUpdate = _onDragUpdate
      ..onEnd = _onDragEnd
      ..onCancel = _onDragCancel;

    _tapGestureRecognizer = TapGestureRecognizer()
      ..team = team
      ..onTapDown = _onTapDown
      ..onTapUp = _onTapUp;

    _overlayStartAnimation = CurvedAnimation(
        parent: _state.overlayStartController, curve: Curves.fastOutSlowIn);

    _overlayEndAnimation = CurvedAnimation(
        parent: _state.overlayEndController, curve: Curves.fastOutSlowIn);

    _stateAnimation = CurvedAnimation(
        parent: _state.stateController, curve: Curves.easeInOut);

    _trackOffset = _getTrackOffset();

    if (isDateTime) {
      _minInms = min.millisecondsSinceEpoch.toDouble();
      _maxInms = max.millisecondsSinceEpoch.toDouble();
      _valuesInms = SfRangeValues(
          _values.start.millisecondsSinceEpoch.toDouble(),
          _values.end.millisecondsSinceEpoch.toDouble());
    }

    _visibleLabels = <String>[];
    _majorTickPositions = <double>[];
    _minorTickPositions = <double>[];
    _updateTextPainter();
  }

  final _SfRangeSelectorState _state;

  final TextPainter _textPainter = TextPainter();

  final double _minTrackWidth = kMinInteractiveDimension * 3;

  Offset _trackOffset;

  double _labelsCount;

  double _minInms;

  double _maxInms;

  SfThumb _activeThumb;

  SfRangeValues _valuesInms;

  bool _showThumbStroke = false;

  bool _willDrawTooltip = false;

  bool _isInteractionEnd = true;

  List<String> _visibleLabels;

  List<double> _majorTickPositions;

  List<double> _minorTickPositions;

  Animation<double> _overlayStartAnimation;

  Animation<double> _overlayEndAnimation;

  Animation<double> _stateAnimation;

  HorizontalDragGestureRecognizer _dragGestureRecognizer;

  TapGestureRecognizer _tapGestureRecognizer;

  bool isLocked = false;

  bool isDragStart = false;

  bool get isDiscrete => stepSize != null && stepSize > 0;

  Rect get trackRect => _trackShape.getPreferredRect(
      this, _sliderThemeData, Offset.zero, isEnabled);

  Size get actualDivisorSize =>
      _divisorShape.getPreferredSize(_sliderThemeData, isEnabled);

  Size get actualTickSize =>
      _tickShape.getPreferredSize(_sliderThemeData, isEnabled);

  Size get actualMinorTickSize =>
      _minorTickShape.getPreferredSize(_sliderThemeData, isEnabled);

  Size get actualLabelSize => Size.fromHeight(math.max(
      _sliderThemeData.inactiveLabelStyle.fontSize,
      _sliderThemeData.activeLabelStyle.fontSize));

  Size get actualThumbSize =>
      _thumbShape.getPreferredSize(_sliderThemeData, isEnabled);

  Size get actualOverlaySize =>
      _overlayShape.getPreferredSize(_sliderThemeData, isEnabled);

  bool get isDateTime =>
      _min.runtimeType == DateTime &&
      _max.runtimeType == DateTime &&
      _values.start.runtimeType == DateTime &&
      _values.end.runtimeType == DateTime;

  double get _actualTickHeight => _showTicks
      ? _sliderThemeData.tickSize.height +
          (_sliderThemeData.tickOffset != null
              ? _sliderThemeData.tickOffset.dy
              : 0)
      : 0;

  double get _actualMinorTickHeight =>
      _minorTicksPerInterval != null ? actualMinorTickSize.height : 0;

  double get _actualLabelHeight => _showLabels
      ? actualLabelSize.height * _textPainter.textScaleFactor +
          (_sliderThemeData.labelOffset != null
              ? _sliderThemeData.labelOffset.dy
              : 0)
      : 0;

  // FIXME: Renaming required.
  double get elementsActualHeight => math.max(
      2 * _trackOffset.dy,
      _trackOffset.dy +
          _sliderThemeData.trackHeight / 2 +
          math.max(_actualTickHeight, _actualMinorTickHeight) +
          _actualLabelHeight);

  double get minThumbGap =>
      (actualMax - actualMin) * (8 / trackRect.width).clamp(0.0, 1.0);

  double get actualMin => isDateTime ? _minInms : _min;

  double get actualMax => isDateTime ? _maxInms : _max;

  SfRangeValues get actualValues => isDateTime ? _valuesInms : _values;

  dynamic get min => _min;
  dynamic _min;
  set min(dynamic value) {
    if (_min == value) {
      return;
    }
    _min = value;

    if (_min is DateTime) {
      _minInms = _min.millisecondsSinceEpoch.toDouble();
    }
    markNeedsPaint();
  }

  dynamic get max => _max;
  dynamic _max;
  set max(dynamic value) {
    if (_max == value) {
      return;
    }
    _max = value;

    if (_max is DateTime) {
      _maxInms = _max.millisecondsSinceEpoch.toDouble();
    }
    markNeedsPaint();
  }

  SfRangeValues get values => _values;
  SfRangeValues _values;
  set values(SfRangeValues values) {
    if (isDiscrete) {
      if (_activeThumb == SfThumb.start) {
        final dynamic startThumb = _getActualValue(value: values.start);
        values = SfRangeValues(startThumb, values.end);
      } else {
        final dynamic endThumb = _getActualValue(value: values.end);
        values = SfRangeValues(values.start, endThumb);
      }
    }

    if (_values == values) {
      return;
    }

    _values = values;
    if (isDateTime) {
      _valuesInms = SfRangeValues(
          _values.start.millisecondsSinceEpoch.toDouble(),
          _values.end.millisecondsSinceEpoch.toDouble());
    }
  }

  bool get isEnabled => _isEnabled;
  bool _isEnabled;
  set isEnabled(bool value) {
    if (_isEnabled == value) {
      return;
    }
    final bool wasEnabled = isEnabled;
    _isEnabled = value;
    if (wasEnabled != isEnabled) {
      if (isEnabled) {
        _state.stateController.forward();
      } else {
        _state.stateController.reverse();
      }
    }
    markNeedsLayout();
  }

  double get interval => _interval;
  double _interval;
  set interval(double value) {
    if (_interval == value) {
      return;
    }
    _interval = value;
    markNeedsPaint();
  }

  double get stepSize => _stepSize;
  double _stepSize;
  set stepSize(double value) {
    if (_stepSize == value) {
      return;
    }
    _stepSize = value;
    markNeedsPaint();
  }

  int get minorTicksPerInterval => _minorTicksPerInterval;
  int _minorTicksPerInterval;
  set minorTicksPerInterval(int value) {
    if (_minorTicksPerInterval == value) {
      return;
    }
    _minorTicksPerInterval = value;
    markNeedsPaint();
  }

  bool get showTicks => _showTicks;
  bool _showTicks;
  set showTicks(bool value) {
    if (_showTicks == value) {
      return;
    }
    _showTicks = value;
    markNeedsLayout();
  }

  bool get showLabels => _showLabels;
  bool _showLabels;
  set showLabels(bool value) {
    if (_showLabels == value) {
      return;
    }
    _showLabels = value;
    markNeedsLayout();
  }

  bool get showDivisors => _showDivisors;
  bool _showDivisors;
  set showDivisors(bool value) {
    if (_showDivisors == value) {
      return;
    }
    _showDivisors = value;
    markNeedsLayout();
  }

  bool get showTooltip => _showTooltip;
  bool _showTooltip;
  set showTooltip(bool value) {
    if (_showTooltip == value) {
      return;
    }
    _showTooltip = value;
  }

  bool get lockRange => _lockRange;
  bool _lockRange;
  set lockRange(bool value) {
    if (_lockRange == value) {
      return;
    }
    _lockRange = value;
  }

  Color get inactiveColor => _inactiveColor;
  Color _inactiveColor;
  set inactiveColor(Color value) {
    if (_inactiveColor == value) {
      return;
    }
    _inactiveColor = value;
    markNeedsPaint();
  }

  Color get activeColor => _activeColor;
  Color _activeColor;
  set activeColor(Color value) {
    if (_activeColor == value) {
      return;
    }
    _activeColor = value;
    markNeedsPaint();
  }

  LabelPlacement get labelPlacement => _labelPlacement;
  LabelPlacement _labelPlacement;
  set labelPlacement(LabelPlacement value) {
    if (_labelPlacement == value) {
      return;
    }
    _labelPlacement = value;
    markNeedsPaint();
  }

  NumberFormat get numberFormat => _numberFormat;
  NumberFormat _numberFormat;
  set numberFormat(NumberFormat value) {
    if (_numberFormat == value) {
      return;
    }
    _numberFormat = value;
    markNeedsPaint();
  }

  DateIntervalType get dateIntervalType => _dateIntervalType;
  DateIntervalType _dateIntervalType;
  set dateIntervalType(DateIntervalType value) {
    if (_dateIntervalType == value) {
      return;
    }
    _dateIntervalType = value;
    markNeedsPaint();
  }

  DateFormat get dateFormat => _dateFormat;
  DateFormat _dateFormat;
  set dateFormat(DateFormat value) {
    if (_dateFormat == value) {
      return;
    }
    _dateFormat = value;
    markNeedsPaint();
  }

  LabelFormatterCallback get labelFormatterCallback => _labelFormatterCallback;
  LabelFormatterCallback _labelFormatterCallback;
  set labelFormatterCallback(LabelFormatterCallback value) {
    if (_labelFormatterCallback == value) {
      return;
    }
    _labelFormatterCallback = value;
    markNeedsSemanticsUpdate();
  }

  TooltipTextFormatterCallback get tooltipTextFormatterCallback =>
      _tooltipTextFormatterCallback;
  TooltipTextFormatterCallback _tooltipTextFormatterCallback;
  set tooltipTextFormatterCallback(TooltipTextFormatterCallback value) {
    if (_tooltipTextFormatterCallback == value) {
      return;
    }
    _tooltipTextFormatterCallback = value;
    markNeedsSemanticsUpdate();
  }

  SfDivisorShape get divisorShape => _divisorShape;
  SfDivisorShape _divisorShape;
  set divisorShape(SfDivisorShape value) {
    if (_divisorShape == value) {
      return;
    }
    _divisorShape = value;
    markNeedsLayout();
  }

  SfTrackShape get trackShape => _trackShape;
  SfTrackShape _trackShape;
  set trackShape(SfTrackShape value) {
    if (_trackShape == value) {
      return;
    }
    _trackShape = value;
    markNeedsLayout();
  }

  SfTickShape get tickShape => _tickShape;
  SfTickShape _tickShape;
  set tickShape(SfTickShape value) {
    if (_tickShape == value) {
      return;
    }
    _tickShape = value;
    markNeedsLayout();
  }

  SfMinorTickShape get minorTickShape => _minorTickShape;
  SfMinorTickShape _minorTickShape;
  set minorTickShape(SfMinorTickShape value) {
    if (_minorTickShape == value) {
      return;
    }
    _minorTickShape = value;
    markNeedsLayout();
  }

  SfOverlayShape get overlayShape => _overlayShape;
  SfOverlayShape _overlayShape;
  set overlayShape(SfOverlayShape value) {
    if (_overlayShape == value) {
      return;
    }
    _overlayShape = value;
    markNeedsLayout();
  }

  SfThumbShape get thumbShape => _thumbShape;
  SfThumbShape _thumbShape;
  set thumbShape(SfThumbShape value) {
    if (_thumbShape == value) {
      return;
    }
    _thumbShape = value;
    markNeedsLayout();
  }

  SfRangeSliderThemeData get sliderThemeData => _sliderThemeData;
  SfRangeSliderThemeData _sliderThemeData;
  set sliderThemeData(SfRangeSliderThemeData value) {
    if (_sliderThemeData == value) {
      return;
    }
    _sliderThemeData = value;
    markNeedsPaint();
  }

  TextDirection get textDirection => _textDirection;
  TextDirection _textDirection;
  set textDirection(TextDirection value) {
    if (_textDirection == value) {
      return;
    }
    _textDirection = value;
    _updateTextPainter();
    markNeedsPaint();
  }

  MediaQueryData get mediaQueryData => _mediaQueryData;
  MediaQueryData _mediaQueryData;
  set mediaQueryData(MediaQueryData value) {
    if (_mediaQueryData == value) {
      return;
    }
    _mediaQueryData = value;
    _updateTextPainter();
    markNeedsLayout();
  }

  Offset _getTrackOffset() {
    final double dx = <double>[
          actualOverlaySize.width,
          actualThumbSize.width,
          actualTickSize.width,
          actualMinorTickSize.width
        ].reduce(math.max) /
        2;

    final double dy = <double>[
          actualOverlaySize.height,
          actualThumbSize.height,
          actualDivisorSize.height,
          _sliderThemeData.trackHeight
        ].reduce(math.max) /
        2;

    return Offset(dx, dy);
  }

  void _onTapDown(TapDownDetails details) {
    _beginInteraction(globalToLocal(details.globalPosition).dx);
  }

  void _onTapUp(TapUpDetails details) {
    _endInteraction();
  }

  void _onDragStart(DragStartDetails details) {
    isDragStart = true;
    _beginInteraction(globalToLocal(details.globalPosition).dx);
  }

  void _onDragUpdate(DragUpdateDetails details) {
    _isInteractionEnd = false;
    _updateRangeValues(globalToLocal(details.globalPosition).dx,
        deltaX: details.delta.dx);
    markNeedsPaint();
  }

  void _onDragEnd(DragEndDetails details) {
    _endInteraction();
  }

  void _onDragCancel() {
    _endInteraction();
  }

  double _valueToPosition(double value) {
    return _valueToFactor(value) * trackRect.width + trackRect.left;
  }

  double _positionToFactor(double position) {
    final double factor =
        ((position - trackRect.left) / trackRect.width).clamp(0.0, 1.0);
    return (_textDirection == TextDirection.rtl) ? 1.0 - factor : factor;
  }

  double _valueToFactor(dynamic value) {
    final double factor =
        (value == null) ? 0.0 : ((value - actualMin) / (actualMax - actualMin));
    return (_textDirection == TextDirection.rtl) ? 1.0 - factor : factor;
  }

  double _factorToValue(double factor) {
    return factor * (actualMax - actualMin) + actualMin;
  }

  void _beginInteraction(double xPosition) {
    _willDrawTooltip = true;
    _isInteractionEnd = false;

    final double startPosition = _valueToPosition(actualValues.start);
    final double endPosition = _valueToPosition(actualValues.end);
    final double leftThumbWidth = (startPosition - xPosition).abs();
    final double rightThumbWidth = (endPosition - xPosition).abs();

    if (_lockRange &&
        startPosition < (xPosition - _minPreferredTouchWidth) &&
        (xPosition + _minPreferredTouchWidth) < endPosition) {
      if (isDragStart)
        isLocked = true;
      else
        return;
    } else if (rightThumbWidth > leftThumbWidth) {
      _activeThumb = SfThumb.start;
      _state.overlayStartController.forward();
    } else {
      _activeThumb = SfThumb.end;
      _state.overlayEndController.forward();
    }

    _updateRangeValues(xPosition);
  }

  void _endInteraction() {
    if (!_isInteractionEnd) {
      _state.overlayStartController.reverse();
      _state.overlayEndController.reverse();
      _willDrawTooltip = false;
      _isInteractionEnd = true;
      isLocked = false;
      isDragStart = false;
      markNeedsPaint();
    }
  }

  void _updateTextPainter() {
    _textPainter
      ..textDirection = _textDirection
      ..textScaleFactor = _mediaQueryData.textScaleFactor;
  }

  dynamic _getValueFromPosition(double position) {
    double valueFactor =
        ((position - trackRect.left) / trackRect.width).clamp(0.0, 1.0);
    if (_textDirection == TextDirection.rtl) {
      valueFactor = 1.0 - valueFactor;
    }

    final dynamic actualValue = _factorToValue(valueFactor);
    if (isDateTime) {
      return _getActualValue(valueInms: actualValue);
    } else {
      return _getActualValue(value: actualValue);
    }
  }

  void _updateRangeValues(double xPosition, {double deltaX}) {
    double start;
    double end;
    SfRangeValues newValues;
    double startPosition = _valueToPosition(actualValues.start);
    double endPosition = _valueToPosition(actualValues.end);

    if (isLocked) {
      final double lockedRangeWidth = endPosition - startPosition;
      startPosition += deltaX ?? 0;
      endPosition += deltaX ?? 0;
      final double actualMinInPx = _valueToPosition(actualMin);
      final double actualMaxInPx = _valueToPosition(actualMax);
      if (startPosition < actualMinInPx) {
        startPosition = actualMinInPx;
        endPosition = startPosition + lockedRangeWidth;
      } else if (endPosition > actualMaxInPx) {
        endPosition = actualMaxInPx;
        startPosition = endPosition - lockedRangeWidth;
      }
      newValues = SfRangeValues(_getValueFromPosition(startPosition),
          _getValueFromPosition(endPosition));
    } else {
      final double factor = _positionToFactor(xPosition);
      if (isDateTime) {
        start = values.start.millisecondsSinceEpoch.toDouble();
        end = values.end.millisecondsSinceEpoch.toDouble();
      } else {
        start = values.start;
        end = values.end;
      }

      final double value = lerpDouble(actualMin, actualMax, factor);
      if (_activeThumb == SfThumb.start) {
        newValues = values._copyWith(start: math.min(value, end - minThumbGap));
      } else {
        newValues = values._copyWith(end: math.max(value, start + minThumbGap));
      }
      _showThumbStroke = (_valueToFactor(start) - _valueToFactor(end)).abs() *
              trackRect.width <
          actualThumbSize.height;
    }

    _updateNewValues(newValues);
  }

  void _updateNewValues(SfRangeValues newValues) {
    if (_isEnabled) {
      if (_state.widget.onChanged != null &&
          (newValues.start != _values.start || newValues.end != _values.end)) {
        _state.widget.onChanged(newValues);
      }

      if (_state.widget.controller != null) {
        _state.widget.controller.start = newValues.start;
        _state.widget.controller.end = newValues.end;
      } else {
        values = newValues;
        markNeedsPaint();
      }
    }
  }

  dynamic _getActualValue({dynamic value, double valueInms}) {
    if (isDiscrete) {
      if (!isDateTime) {
        final double minMaxDiff = _max - _min;
        value = _factorToValue(
            ((_valueToFactor(value) * (minMaxDiff / stepSize)).round() /
                    (minMaxDiff / stepSize))
                .clamp(0.0, 1.0));
      }
    }

    return isDateTime
        ? (value ?? DateTime.fromMillisecondsSinceEpoch(valueInms.toInt()))
        : value;
  }

  void _generateVisibleLabels() {
    _visibleLabels.clear();
    _majorTickPositions.clear();
    _minorTickPositions.clear();

    String label;
    dynamic currentValue = _min;
    _labelsCount = (isDateTime
                ? _getDateTimeDifference(_min, _max, _dateIntervalType)
                : _max - _min)
            .toDouble() /
        _interval;
    for (int i = 0; i <= _labelsCount; i++) {
      label = _labelFormatterCallback(
          currentValue, _getFormattedText(currentValue));
      _visibleLabels.add(label);

      final double labelPosition = _valueToFactor(
              isDateTime ? currentValue.millisecondsSinceEpoch : currentValue) *
          (trackRect.width);
      if (!_majorTickPositions.contains(labelPosition)) {
        _majorTickPositions.add(labelPosition);
      }
      currentValue = isDateTime
          ? _getNextDate(currentValue, _dateIntervalType, _interval)
          : currentValue + _interval;
    }
    final int majorTicksCount = _majorTickPositions.length;
    if (_minorTicksPerInterval > 0) {
      if (isDateTime) {
        double minorTickPosition;
        DateTime nextDate = _getNextDate(_min, _dateIntervalType, _interval);
        DateTime currentActualDate =
            _getNextDate(nextDate, _dateIntervalType, -_interval);
        for (int i = 1; i <= majorTicksCount; i++) {
          // Need to divide the region based on _minorTicksPerInterval.
          // So, added 1 with _minorTicksPerInterval.
          final double intervalDiff = _getDateTimeDifference(
                  currentActualDate, nextDate, _dateIntervalType) /
              (_minorTicksPerInterval + 1);
          // On dividing region equally between two dates, then it will be equal to 1.0.
          if (intervalDiff == 1.0) {
            // To get nextDate of minorTick, need to pass intervals.
            // So, iteration value is used as interval and is started with 1.
            for (double j = 1; j <= _minorTicksPerInterval; j++) {
              final DateTime nextMinorDate =
                  _getNextDate(currentActualDate, _dateIntervalType, j);
              minorTickPosition = _valueToPosition(
                      nextMinorDate.millisecondsSinceEpoch.toDouble()) -
                  trackRect.left;
              _minorTickPositions.add(minorTickPosition);
            }
          } else {
            final double minorPositionDiff = (nextDate.millisecondsSinceEpoch -
                    currentActualDate.millisecondsSinceEpoch) /
                (_minorTicksPerInterval + 1);
            for (int j = 1; j <= _minorTicksPerInterval; j++) {
              minorTickPosition = _valueToPosition(
                      currentActualDate.millisecondsSinceEpoch +
                          (j * minorPositionDiff)) -
                  trackRect.left;
              _minorTickPositions.add(minorTickPosition);
            }
          }
          currentActualDate = nextDate;
          nextDate =
              _getNextDate(currentActualDate, _dateIntervalType, interval);
        }
      } else {
        for (int i = 0; i <= majorTicksCount - 1; i++) {
          final double minorPositionDiff = ((i + 1 < majorTicksCount
                      ? _majorTickPositions[i + 1]
                      : trackRect.width) -
                  _majorTickPositions[i]) /
              (_minorTicksPerInterval + 1);
          for (int j = 1; j <= _minorTicksPerInterval; j++) {
            _minorTickPositions
                .add(_majorTickPositions[i] + j * minorPositionDiff);
          }
        }
      }
    }
  }

  void _onRangeControllerListener() {
    if (_state.mounted &&
        _state.widget.controller != null &&
        (_values.start != _state.widget.controller.start ||
            _values.end != _state.widget.controller.end)) {
      values = SfRangeValues(
          _getActualValue(value: _state.widget.controller.start),
          _getActualValue(value: _state.widget.controller.end));
      markNeedsPaint();
    }
  }

  @override
  void systemFontsDidChange() {
    super.systemFontsDidChange();
    _textPainter.markNeedsLayout();
    _updateTextPainter();
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    _overlayStartAnimation?.addListener(markNeedsPaint);
    _overlayEndAnimation?.addListener(markNeedsPaint);
    _stateAnimation?.addListener(markNeedsPaint);
    if (_state.widget.controller != null) {
      _state.widget.controller.addListener(_onRangeControllerListener);
    }
  }

  @override
  void detach() {
    _overlayStartAnimation?.removeListener(markNeedsPaint);
    _overlayEndAnimation?.removeListener(markNeedsPaint);
    _stateAnimation?.removeListener(markNeedsPaint);
    if (_state.widget.controller != null) {
      _state.widget.controller.removeListener(_onRangeControllerListener);
    }

    super.detach();
  }

  @override
  bool hitTestSelf(Offset position) => isEnabled;

  @override
  bool hitTestChildren(BoxHitTestResult result, {Offset position}) => false;

  @override
  void setupParentData(RenderObject child) {
    if (child.parentData is! BoxParentData) {
      child.parentData = BoxParentData();
    }
  }

  @override
  void performLayout() {
    double childHeight = 0.0;
    double childWidth = 0.0;
    final double bottomEdgeInset = math.max(
        _trackOffset.dy + _sliderThemeData.trackHeight / 2,
        _sliderThemeData.trackHeight +
            math.max(_actualTickHeight, _actualMinorTickHeight) +
            _actualLabelHeight);
    if (child != null) {
      final double maxRadius = _trackOffset.dx;
      final BoxConstraints childConstraints = constraints.deflate(
          EdgeInsets.only(
              left: maxRadius, right: maxRadius, bottom: bottomEdgeInset));
      child.layout(childConstraints, parentUsesSize: true);
      final BoxParentData childParentData = child.parentData;
      childParentData.offset = Offset(maxRadius, 0);
      childHeight = child.size.height;
      childWidth = child.size.width;
    }
    final double actualWidth = childWidth > 0.0
        ? (childWidth + 2 * _trackOffset.dx)
        : _minTrackWidth + 2 * _trackOffset.dx;

    final double actualHeight = childHeight +
        elementsActualHeight +
        _sliderThemeData.trackHeight / 2 -
        (childHeight > _trackOffset.dy ? _trackOffset.dy : 0.0);

    size = Size(
        constraints.hasBoundedWidth && (constraints.maxWidth < actualWidth)
            ? constraints.maxWidth
            : actualWidth,
        constraints.hasBoundedHeight && (constraints.maxHeight < actualHeight)
            ? constraints.maxHeight
            : actualHeight);

    if (interval != null && interval > 0) {
      _generateVisibleLabels();
    }
  }

  @override
  void handleEvent(PointerEvent event, HitTestEntry entry) {
    // Checked [_isInteractionEnd] for avoiding multi touch.
    if (_isInteractionEnd && event.down && event is PointerDownEvent) {
      _dragGestureRecognizer.addPointer(event);
      _tapGestureRecognizer.addPointer(event);
    }
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    double childHeight = 0.0;
    if (child != null) {
      final BoxParentData childParentData = child.parentData;
      context.paintChild(child, childParentData.offset + offset);
      childHeight = child.size.height;
      if (childHeight >= constraints.maxHeight) {
        childHeight -= elementsActualHeight -
            math.max(actualOverlaySize.height, actualThumbSize.height) / 2;
      }
    }

    final Offset actualTrackOffset = Offset(
        offset.dx,
        offset.dy +
            math.max(childHeight,
                _trackOffset.dy - _sliderThemeData.trackHeight / 2));

    // Drawing track.
    final Rect _trackRect = _trackShape.getPreferredRect(
        this, _sliderThemeData, actualTrackOffset, _isEnabled);
    final dynamic thumbStartPosition =
        _valueToFactor(actualValues.start) * _trackRect.width;
    final dynamic thumbEndPosition =
        _valueToFactor(actualValues.end) * _trackRect.width;
    final Offset startThumbCenter =
        Offset(_trackRect.left + thumbStartPosition, _trackRect.center.dy);
    final Offset endThumbCenter =
        Offset(_trackRect.left + thumbEndPosition, _trackRect.center.dy);

    _trackShape.paint(
        context, actualTrackOffset, startThumbCenter, endThumbCenter,
        isEnabled: _isEnabled,
        parentBox: this,
        themeData: _sliderThemeData,
        animation: _stateAnimation,
        textDirection: _textDirection);

    if (_interval != null && _interval > 0) {
      int _dateTimePos = 0;
      final double dx = _trackRect.left;
      final double dy = _trackRect.top;
      final double halfTrackHeight = _trackRect.height / 2;
      final double divisorRadius =
          _divisorShape.getPreferredSize(_sliderThemeData, _isEnabled).width /
              2;
      final double tickRadius =
          _tickShape.getPreferredSize(_sliderThemeData, _isEnabled).width / 2;

      double textvalue = isDateTime ? 0.0 : _min;
      int minorTickIndex = 0;
      final double maxRange = isDateTime ? _labelsCount : _max;

      while (textvalue <= maxRange) {
        final double tickPosition = _majorTickPositions[_dateTimePos];

        // Drawing divisors.
        if (showDivisors) {
          Offset divisorCenter =
              Offset(dx + tickPosition, dy + halfTrackHeight);
          if (_majorTickPositions[_dateTimePos] == 0.0)
            divisorCenter =
                Offset(dx + tickPosition + divisorRadius, dy + halfTrackHeight);
          else if (_majorTickPositions[_dateTimePos] == _trackRect.width)
            divisorCenter =
                Offset(dx + tickPosition - divisorRadius, dy + halfTrackHeight);
          _divisorShape.paint(
              context, divisorCenter, startThumbCenter, endThumbCenter,
              isEnabled: _isEnabled,
              parentBox: this,
              themeData: _sliderThemeData,
              animation: _stateAnimation,
              textDirection: _textDirection);
        }

        // Drawing ticks.
        if (showTicks) {
          Offset actualTickOffset =
              Offset(dx + tickPosition, dy + _trackRect.height) +
                  (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          if (_majorTickPositions[_dateTimePos] == 0.0)
            actualTickOffset =
                Offset(dx + tickPosition + tickRadius, dy + _trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          else if (_majorTickPositions[_dateTimePos] == _trackRect.width)
            actualTickOffset =
                Offset(dx + tickPosition - tickRadius, dy + _trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          _tickShape.paint(
              context, actualTickOffset, startThumbCenter, endThumbCenter,
              isEnabled: _isEnabled,
              parentBox: this,
              themeData: _sliderThemeData,
              animation: _stateAnimation,
              textDirection: _textDirection);
        }

        if (_minorTicksPerInterval > 0) {
          for (int j = 0; j < _minorTicksPerInterval; j++) {
            final double currentMinorTickPosition =
                _minorTickPositions[minorTickIndex];
            minorTickIndex++;

            if (currentMinorTickPosition < trackRect.width &&
                currentMinorTickPosition > 0) {
              final Offset actualTickOffset = Offset(
                      dx + currentMinorTickPosition, dy + _trackRect.height) +
                  (_sliderThemeData.tickOffset ?? const Offset(0, 0));
              _minorTickShape.paint(
                  context, actualTickOffset, startThumbCenter, endThumbCenter,
                  isEnabled: _isEnabled,
                  parentBox: this,
                  themeData: _sliderThemeData,
                  animation: _stateAnimation,
                  textDirection: _textDirection);
            }
          }
        }

        // Drawing label.
        if (showLabels) {
          final dynamic labelText = _visibleLabels[_dateTimePos];
          final bool isRTL = _textDirection == TextDirection.rtl;
          double offsetX = dx + tickPosition;
          if (labelPlacement == LabelPlacement.betweenTicks) {
            offsetX += ((_dateTimePos + 1 <= _labelsCount
                        ? _majorTickPositions[_dateTimePos + 1]
                        : (isRTL ? _trackRect.left : _trackRect.width)) -
                    tickPosition) /
                2;
            if (isRTL
                ? offsetX <= _trackRect.left
                : offsetX - dx >= _trackRect.width) {
              break;
            }
          }
          final Offset actualLabelOffset = Offset(
                offsetX,
                dy + _trackRect.height + _actualTickHeight,
              ) +
              (_sliderThemeData.labelOffset ?? const Offset(0, 0));

          _drawText(context, actualLabelOffset, startThumbCenter, labelText,
              endThumbCenter: endThumbCenter,
              parentBox: this,
              themeData: _sliderThemeData,
              animation: _stateAnimation,
              textPainter: _textPainter,
              textDirection: _textDirection);
        }

        final dynamic intervalDiff = isDateTime ? 1 : interval;
        textvalue += intervalDiff;
        _dateTimePos += 1;
      }
    }

    // Drawing selected and unselected region rects.
    final Paint paint = Paint()
      ..isAntiAlias = true
      ..color = sliderThemeData.inactiveRegionColor;
    if (child != null && child.size.height > 1 && child.size.width > 1) {
      context.canvas.drawRect(
          Rect.fromLTRB(
              _trackRect.left, offset.dy, startThumbCenter.dx, _trackRect.top),
          paint);
      paint.color = sliderThemeData.activeRegionColor;
      context.canvas.drawRect(
          Rect.fromLTRB(startThumbCenter.dx, offset.dy, endThumbCenter.dx,
              _trackRect.top),
          paint);
      paint.color = sliderThemeData.inactiveRegionColor;
      context.canvas.drawRect(
          Rect.fromLTRB(
              endThumbCenter.dx, offset.dy, _trackRect.right, _trackRect.top),
          paint);
    }

    final bool isLeftThumbActive = _activeThumb == SfThumb.start;
    Offset thumbCenter = isLeftThumbActive ? endThumbCenter : startThumbCenter;
    // Drawing thumb.
    _thumbShape.paint(context, thumbCenter,
        isEnabled: _isEnabled,
        parentBox: this,
        themeData: _sliderThemeData,
        animation: _stateAnimation,
        textDirection: _textDirection,
        thumb: isLeftThumbActive ? SfThumb.end : SfThumb.start);

    thumbCenter = isLeftThumbActive ? startThumbCenter : endThumbCenter;
    // Drawing overlay.
    _overlayShape.paint(context, thumbCenter,
        isEnabled: _isEnabled,
        parentBox: this,
        themeData: _sliderThemeData,
        animation:
            isLeftThumbActive ? _overlayStartAnimation : _overlayEndAnimation,
        thumb: _activeThumb);

    // Drawing thumb.
    _thumbShape.paint(context, thumbCenter,
        isEnabled: _isEnabled,
        parentBox: this,
        themeData: _sliderThemeData,
        animation: _stateAnimation,
        textDirection: _textDirection,
        thumb: _activeThumb);

    if (_thumbShape is _SfThumbShape && _showThumbStroke) {
      paint
        ..color = Colors.white
        ..style = PaintingStyle.stroke
        ..strokeWidth = 0.5;

      if (_activeThumb == SfThumb.start) {
        context.canvas
            .drawCircle(startThumbCenter, sliderThemeData.thumbRadius, paint);
      } else {
        context.canvas
            .drawCircle(endThumbCenter, sliderThemeData.thumbRadius, paint);
      }
    }

    if (_showTooltip && _willDrawTooltip) {
      paint
        ..color = _sliderThemeData.tooltipBackgroundColor
        ..style = PaintingStyle.fill
        ..strokeWidth = 0;
      _drawTooltip(
          context,
          offset,
          Offset(actualTrackOffset.dx + _trackOffset.dx, actualTrackOffset.dy),
          startThumbCenter,
          endThumbCenter,
          paint,
          _trackRect);
    }
  }

  String _getFormattedText(dynamic value) {
    if (isDateTime) {
      return _dateFormat != null ? _dateFormat.format(value) : value.toString();
    }
    return _numberFormat.format(value);
  }

  // Here 10 is a gap between tooltip nose and thumb.
  double get tooltipStartY => actualThumbSize.height / 2 + 10;

  void _drawTooltip(
      PaintingContext context,
      Offset offset,
      Offset trackStartOffset,
      Offset startThumbCenter,
      Offset endThumbCenter,
      Paint paint,
      Rect trackRect) {
    Offset thumbCenter =
        _activeThumb == SfThumb.start ? endThumbCenter : startThumbCenter;
    dynamic actualText = _getValueFromPosition(thumbCenter.dx - offset.dx);
    String tooltipText =
        tooltipTextFormatterCallback(actualText, _getFormattedText(actualText));
    TextSpan textSpan =
        TextSpan(text: tooltipText, style: _sliderThemeData.tooltipTextStyle);
    _textPainter.text = textSpan;
    _textPainter.layout();
    Path path = getTooltipPath(_textPainter.size + _tooltipTextPadding,
        tooltipStartY, trackStartOffset, trackRect, thumbCenter.dx, size);
    context.canvas
        .drawPath(path, paint..color = _sliderThemeData.tooltipBackgroundColor);

    final double leftPadding = _tooltipTextPadding.dx / 2;
    final double minLeftdx = trackStartOffset.dx + leftPadding;
    double dx = _getActualTooltipTextX(
        minLeftdx, thumbCenter, _textPainter.width / 2, leftPadding, trackRect);
    final double dy = tooltipStartY +
        _tooltipTriangleHeight +
        (path.getBounds().size.height - _tooltipTriangleHeight) / 2 +
        _textPainter.height / 2 +
        trackRect.height / 2;
    _textPainter.paint(context.canvas, Offset(dx, thumbCenter.dy - dy));

    thumbCenter =
        _activeThumb == SfThumb.start ? startThumbCenter : endThumbCenter;
    actualText = _getValueFromPosition(thumbCenter.dx - offset.dx);
    tooltipText =
        tooltipTextFormatterCallback(actualText, _getFormattedText(actualText));
    textSpan =
        TextSpan(text: tooltipText, style: _sliderThemeData.tooltipTextStyle);
    _textPainter.text = textSpan;
    _textPainter.layout();
    path = getTooltipPath(_textPainter.size + _tooltipTextPadding,
        tooltipStartY, trackStartOffset, trackRect, thumbCenter.dx, size);
    context.canvas.drawPath(path, paint);

    dx = _getActualTooltipTextX(
        minLeftdx, thumbCenter, _textPainter.width / 2, leftPadding, trackRect);
    _textPainter.paint(context.canvas, Offset(dx, thumbCenter.dy - dy));
  }

  double _getActualTooltipTextX(double minLeftdx, Offset thumbCenter,
      double halfTextWidth, double leftPadding, Rect trackRect) {
    final double dx = thumbCenter.dx - halfTextWidth;
    return dx >= minLeftdx
        ? thumbCenter.dx + halfTextWidth > trackRect.right
            ? trackRect.right - _textPainter.width - leftPadding
            : dx
        : minLeftdx;
  }
}
