part of sliders;

/// A Material Design range slider.
///
/// Used to select a range between [min] and [max]. It supports both numeric and date ranges.
///
/// The range slider elements are:
///
/// * The "track", which is the rounded rectangle in which the thumbs are slides over.
/// * The "thumb", which is a shape that slides horizontally when the user drags it.
/// * The "active" side of the range slider is between the left and right thumbs.
/// * The "inactive" side of the range slider is between the [min] value and the
/// left thumb, and the right thumb and the [max] value. For RTL, the inactive side of the range slider is between
/// the [max] value and the left thumb, and the right thumb and the [min] value.
/// * The "divisors", which is a shape that renders on the track based on the given [interval] value.
/// * The "ticks", which is a shape that rendered based on given [interval] value.
/// Basically, it is rendered below the track. It is also called “major ticks”.
/// * The "minor ticks", which is a shape that renders between two major ticks
/// based on given [minorTicksPerInterval] value. Basically, it is rendered below the track.
/// * The "labels", which is a text that rendered based on given [interval] value.
/// Basically, it is rendered below the track and the major ticks.
///
/// The range slider will be disabled if [onChanged] is null or [min] is equal to [max].
///
/// The range slider widget doesn’t maintains any state.
/// Alternatively, the widget calls the [onChanged] callback with the new values
/// when the state of range slider changes. To update the range slider’s visual
/// appearance with the new values, rebuilds the range slider with the new values.
///
/// Range slider can be customized using the [SfRangeSliderTheme](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderTheme-class.html) with the help of [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), or the [SfTheme](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfTheme-class.html) with the help of [SfThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfThemeData-class.html).
/// It is also possible to override the appearance using [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) which were set using the properties of the widget like [activeColor] and [inactiveColor].
///
/// ## Example
///
/// This snippet shows how to create a numeric [SfRangeSlider].
///
/// ```dart
/// SfRangeValues _values = SfRangeValues(4.0, 8.0);
///
/// @override
/// Widget build(BuildContext context) {
///   return MaterialApp(
///      home: Scaffold(
///          body: Center(
///              child: SfRangeSlider(
///                     min: 0.0,
///                     max: 10.0,
///                     values: _values,
///                     onChanged: (SfRangeValues newValues) {
///                         setState(() {
///                             _values = newValues;
///                         });
///                   },
///              )
///           )
///       )
///   );
/// }
/// ```
///
/// This snippet shows how to create a date [SfRangeSlider].
///
/// ```dart
/// SfRangeValues _values = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
///
/// SfRangeSlider(
///  min: DateTime(2000, 01, 01, 00),
///  max: DateTime(2005, 12, 31, 24),
///  values: _values,
///  interval: 1,
///  dateFormat: DateFormat.y(),
/// dateIntervalType: DateIntervalType.years,
///  onChanged: (SfRangeValues newValues) {
///    setState(() {
///      _values = newValues;
///    });
///  },
/// )
/// ```
///
/// See also:
///
/// * [SfTheme](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfTheme-class.html) and [SfThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfThemeData-class.html), for customizing the visual appearance of the range slider.
/// * [numberFormat] and [dateFormat], for formatting the numeric and date labels.
/// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the visual appearance of the range slider.
class SfRangeSlider extends StatefulWidget {
  const SfRangeSlider(
      {Key key,
      this.min = 0.0,
      this.max = 1.0,
      @required this.values,
      @required this.onChanged,
      this.interval,
      this.stepSize,
      this.minorTicksPerInterval,
      this.showTicks = false,
      this.showLabels = false,
      this.showDivisors = false,
      this.showTooltip = false,
      this.inactiveColor,
      this.activeColor,
      this.numberFormat,
      this.dateFormat,
      this.dateIntervalType,
      this.labelPlacement,
      this.labelFormatterCallback,
      this.tooltipTextFormatterCallback,
      this.trackShape,
      this.divisorShape,
      this.overlayShape,
      this.thumbShape,
      this.tickShape,
      this.minorTickShape})
      : assert(min != null),
        assert(max != null),
        assert(interval == null || interval > 0),
        super(key: key);

  /// The minimum value that the user can select.
  ///
  /// Defaults to 0.0. Must be less than the [max].
  final dynamic min;

  /// The maximum value that the user can select.
  ///
  /// Defaults to 1.0. Must be greater than the [min].
  final dynamic max;

  /// The values currently selected in the range slider.
  ///
  /// The range slider's start and end thumbs are drawn corresponding to these values.
  ///
  /// For date values, the range slider doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType], and [dateFormat] for date values,
  /// if the labels, ticks, and divisors are needed.
  ///
  /// This snippet shows how to create a numeric [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// This snippet shows how to create a date [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSlider(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   values: _values,
  ///   interval: 1,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///      _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [onChanged], to update the visual appearance of
  /// the range slider when the user drags the thumb through interaction.
  /// * [showTicks], to render major ticks at given interval.
  /// * [showLabels], to render labels at given interval.
  final SfRangeValues values;

  /// Called when the user is selecting a new values for the slider by dragging.
  ///
  /// The range slider passes the new values to the callback but
  /// does not change its state until the parent widget rebuilds
  /// the range slider with new values.
  ///
  /// If it is null, the range slider will be disabled.
  ///
  /// This snippet shows how to create a numeric [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  ///
  final ValueChanged<SfRangeValues> onChanged;

  /// Splits the range slider into given interval.
  /// It is mandatory if labels, major ticks and divisors are needed.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range slider will render the labels, major ticks,
  /// and divisors at 0.0, 2.0, 4.0 and so on.
  ///
  /// For date values, the range slider doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType],
  /// and [dateFormat] for date values, if labels, ticks, and divisors are needed.
  ///
  /// For example, if [min] is DateTime(2000, 01, 01, 00) and [max] is DateTime(2005, 12, 31, 24),
  /// [interval] is 1.0, [dateFormat] is DateFormat.y(), and [dateIntervalType] is DateIntervalType.years,
  /// then the range slider will render the labels, major ticks, and divisors at 2000, 2001, 2002 and so on.
  ///
  /// Defaults to `null`. Must be greater than 0.
  ///
  /// This snippet shows how to set numeric interval in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 2,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// This snippet shows how to set date interval in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSlider(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   values: _values,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///      _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [showDivisors], to render divisors at given interval.
  /// * [showTicks], to render major ticks at given interval.
  /// * [showLabels], to render labels at given interval.
  final double interval;

  /// Option to select discrete values.
  ///
  /// [stepSize] doesn’t work for [DateTime] range slider.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [stepSize] is 2.0,
  /// the range slider will move the thumbs at 0.0, 2.0, 4.0, 6.0, 8.0 and 10.0.
  ///
  /// Defaults to `null`. Must be greater than 0.
  ///
  /// This snippet shows how to set numeric stepSize in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   stepSize: 2,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  final double stepSize;

  /// Number of smaller ticks between two major ticks.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range slider will render the major ticks at 0.0, 2.0, 4.0 and so on.
  /// If minorTicksPerInterval is 1, then smaller ticks will be rendered on 1.0, 3.0 and so on.
  ///
  /// Defaults to `null`. Must be greater than 0.
  ///
  /// This snippet shows how to show minor ticks in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 2,
  ///   showTicks: true,
  ///   minorTicksPerInterval: 1,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [showTicks], to render major ticks at given interval.
  /// * [minorTickShape] and [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing
  /// the minor tick’s visual appearance.
  final int minorTicksPerInterval;

  /// Option to render the major ticks on the track.
  ///
  /// It is a shape which is used to represent the major interval points of the track.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range slider will render the major ticks at 0.0, 2.0, 4.0 and so on.
  ///
  /// Defaults to `false`.
  ///
  /// This snippet shows how to show major ticks in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 2,
  ///   showTicks: true,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [tickShape] and [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) for customizing the
  /// major tick’s visual appearance.
  final bool showTicks;

  /// Option to render the labels on given interval.
  ///
  /// Defaults to `false`.
  ///
  /// This snippet shows how to show labels in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 1,
  ///   showLabels: true,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [numberFormat] and [dateFormat], for formatting the numeric and date labels.
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the appearance of the labels.
  final bool showLabels;

  /// Option to render the divisors on the track.
  ///
  /// It is a shape which is used to represent the major interval points of the track.
  ///
  /// For example, if [min] is 0.0 and [max] is 10.0 and [interval] is 2.0,
  /// the range slider will render the divisors at 0.0, 2.0, 4.0 and so on.
  ///
  /// Defaults to `false`.
  ///
  /// This snippet shows how to show divisors in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 2,
  ///   showDivisors: true,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [showTicks], to render major ticks at given interval.
  /// * [showLabels], to render labels at given interval.
  /// * [divisorShape] and [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html) for customizing
  /// the divisor’s visual appearance.
  final bool showDivisors;

  /// Option to enable tooltips for both the thumbs.
  ///
  /// Used to clearly indicate the current selection of the ranges during interaction.
  ///
  /// By default, tooltip text is formatted with either [numberFormat] or [dateFormat].
  ///
  /// This snippet shows how to show tooltip in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   showTooltip: true,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [tooltipTextFormatterCallback], for changing the default tooltip text.
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the appearance of the tooltip text.
  final bool showTooltip;

  /// Color applied to the inactive track and active divisors.
  ///
  /// The inactive side of the range slider is between the [min] value and
  /// the left thumb, and the right thumb and the [max] value.
  ///
  /// For RTL, the inactive side of the range slider is between the [max] value and
  /// the left thumb, and the right thumb and the [min] value.
  ///
  /// This snippet shows how to set inactive color in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   inactiveColor: Colors.red,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the individual
  /// inactive range slider element’s visual.
  final Color inactiveColor;

  /// Color applied to the active track, thumb, overlay, and inactive divisors.
  ///
  /// The active side of the range slider is between the start and end thumbs.
  ///
  /// This snippet shows how to set active color in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 1,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   activeColor: Colors.red,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///    },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [SfRangeSliderThemeData](https://pub.dev/documentation/syncfusion_flutter_core/latest/theme/SfRangeSliderThemeData-class.html), for customizing the individual active range slider element’s visual.
  final Color activeColor;

  /// Formats the numeric labels.
  ///
  /// Defaults to `null`.
  ///
  /// This snippet shows how to set number format in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(4.0, 8.0);
  ///
  /// SfRangeSlider(
  ///   min: 0.0,
  ///   max: 10.0,
  ///   values: _values,
  ///   interval: 2,
  ///   showTicks: true,
  ///   showLabels: true,
  ///   numberFormat: NumberFormat(‘\$’),
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [labelFormatterCallback], for formatting the numeric and date labels.
  final NumberFormat numberFormat;

  /// Formats the date labels. It is mandatory for date [SfRangeSlider].
  ///
  /// For date values, the range selector doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType],
  /// and [dateFormat] for date values, if labels, ticks, and divisors are needed.
  ///
  /// Defaults to `null`.
  ///
  /// This snippet shows how to set date format in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSlider(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   values: _values,
  ///   interval: 1,
  ///   showLabels: true,
  ///   showTicks: true,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  ///
  /// See also:
  ///
  /// * [interval], for setting the interval.
  /// * [numberFormat], for formatting the numeric labels.
  /// * [labelFormatterCallback], for formatting the numeric and date label.
  /// * [dateIntervalType], for changing the interval type.
  final DateFormat dateFormat;

  /// The type of date interval. It is mandatory for date [SfRangeSlider].
  ///
  /// It can be years to seconds.
  ///
  /// For date values, the range selector doesn’t have auto interval support.
  /// So, you may need to set [interval], [dateIntervalType],
  /// and [dateFormat] for date values, if labels, ticks, and divisors are needed.
  ///
  /// Defaults to `null`.
  ///
  /// This snippet shows how to set date interval type in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSlider(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   values: _values,
  ///   interval: 1,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  final DateIntervalType dateIntervalType;

  /// Option to place the labels either between the major ticks or on the major ticks.
  ///
  /// Defaults to [LabelPlacement.onTicks].
  ///
  /// This snippet shows how to set label placement in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(DateTime(2002, 01, 01), DateTime(2003, 01, 01));
  ///
  /// SfRangeSlider(
  ///   min: DateTime(2000, 01, 01, 00),
  ///   max: DateTime(2005, 12, 31, 24),
  ///   values: _values,
  ///   interval: 1,
  ///   dateFormat: DateFormat.y(),
  ///   dateIntervalType: DateIntervalType.years,
  ///   labelPlacement: LabelPlacement.betweenTicks,
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  final LabelPlacement labelPlacement;

  /// Signature for formatting or changing the whole numeric or date label text.
  ///
  /// * The actual value without formatting is given by `actualValue`. It is either [DateTime] or [double] based on given [values].
  /// * The formatted value based on the numeric or date format is given by `formattedText`.
  ///
  /// This snippet shows how to set label format in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(100.0, 10000.0);
  ///
  /// SfRangeSlider(
  ///   min: 100.0,
  ///   max: 10000.0,
  ///   values: _values,
  ///   interval: 9000.0,
  ///   showLabels: true,
  ///   showTicks: true,
  ///   labelFormatterCallback: (dynamic actualValue, String formattedText) {
  ///     return actualValue == 10000 ? '\$ $ formattedText +' : '\$ $ formattedText';
  ///   },
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  final LabelFormatterCallback labelFormatterCallback;

  /// Signature for formatting or changing the whole tooltip label text.
  ///
  /// * The actual value without formatting is given by `actualValue`. It is either [DateTime] or [double] based on given [values].
  /// * The formatted value based on the numeric or date format is given by `formattedText`.
  ///
  /// This snippet shows how to set tooltip format in [SfRangeSlider].
  ///
  /// ```dart
  /// SfRangeValues _values = SfRangeValues(DateTime(2010, 01, 01, 13, 00, 00), DateTime(2010, 01, 01, 17, 00, 00));
  ///
  /// SfRangeSlider(
  ///   min: DateTime(2010, 01, 01, 9, 00, 00),
  ///   max: DateTime(2010, 01, 01, 21, 05, 00),
  ///   values: _values,
  ///   interval: 4,
  ///   showLabels: true,
  ///   showTooltip: true,
  ///   dateFormat: DateFormat('h a'),
  ///   dateIntervalType: DateIntervalType.hours,
  ///   tooltipTextFormatterCallback: (dynamic actualValue, String formattedText) {
  ///     return DateFormat('h:mm a').format(actualValue);
  ///   },
  ///   onChanged: (SfRangeValues newValues) {
  ///     setState(() {
  ///       _values = newValues;
  ///     });
  ///   },
  /// )
  /// ```
  final TooltipTextFormatterCallback tooltipTextFormatterCallback;

  /// Base class for [SfRangeSlider] divisors shapes.
  final SfDivisorShape divisorShape;

  /// Base class for [SfRangeSlider] track shapes.
  final SfTrackShape trackShape;

  /// Base class for [SfRangeSlider] major tick shapes.
  final SfTickShape tickShape;

  /// Base class for [SfRangeSlider] minor tick shapes.
  final SfMinorTickShape minorTickShape;

  /// Base class for [SfRangeSlider] overlay shapes.
  final SfOverlayShape overlayShape;

  ///  Base class for [SfRangeSlider] thumb shapes.
  final SfThumbShape thumbShape;

  @override
  _SfRangeSliderState createState() => _SfRangeSliderState();
}

class _SfRangeSliderState extends State<SfRangeSlider>
    with TickerProviderStateMixin {
  AnimationController overlayStartController;
  AnimationController overlayEndController;
  AnimationController stateController;
  RangeController rangeController;
  final Duration duration = const Duration(milliseconds: 100);

  void _onChanged(SfRangeValues values) {
    if (values != widget.values) {
      widget.onChanged(values);
    }
  }

  String _getFormattedLabelText(dynamic actualText, String formattedText) {
    return formattedText;
  }

  String _getFormattedTooltipText(dynamic actualText, String formattedText) {
    return formattedText;
  }

  @override
  void dispose() {
    overlayStartController?.dispose();
    overlayEndController?.dispose();
    stateController?.dispose();

    super.dispose();
  }

  @override
  void initState() {
    super.initState();
    overlayStartController =
        AnimationController(vsync: this, duration: duration);
    overlayEndController = AnimationController(vsync: this, duration: duration);
    stateController = AnimationController(vsync: this, duration: duration);
    stateController.value = widget.onChanged != null ? 1.0 : 0.0;
  }

  @override
  Widget build(BuildContext context) {
    SyncfusionLicense.validateLicense(context);
    final ThemeData themeData = Theme.of(context);
    final bool isActive = widget.onChanged != null;
    SfRangeSliderThemeData sliderThemeData = SfRangeSliderTheme.of(context);
    sliderThemeData = sliderThemeData.copyWith(
        trackHeight: sliderThemeData.trackHeight,
        tickSize: sliderThemeData.tickSize,
        minorTickSize: sliderThemeData.minorTickSize,
        tickOffset: sliderThemeData.tickOffset,
        labelOffset: sliderThemeData.labelOffset ??
            (widget.showTicks
                ? const Offset(0.0, 5.0)
                : const Offset(0.0, 13.0)),
        inactiveLabelStyle: sliderThemeData.inactiveLabelStyle ??
            themeData.textTheme.body2.copyWith(
                color: isActive
                    ? themeData.textTheme.body2.color.withOpacity(0.87)
                    : themeData.colorScheme.onSurface.withOpacity(0.32)),
        activeLabelStyle: sliderThemeData.activeLabelStyle ??
            themeData.textTheme.body2.copyWith(
                color: isActive
                    ? themeData.textTheme.body2.color.withOpacity(0.87)
                    : themeData.colorScheme.onSurface.withOpacity(0.32)),
        tooltipTextStyle: sliderThemeData.tooltipTextStyle ??
            themeData.textTheme.body2.copyWith(
                color: sliderThemeData.brightness == Brightness.light
                    ? const Color.fromRGBO(255, 255, 255, 1)
                    : const Color.fromRGBO(0, 0, 0, 1)),
        inactiveTrackColor: widget.inactiveColor ??
            sliderThemeData.inactiveTrackColor ??
            themeData.primaryColor.withOpacity(0.24),
        activeTrackColor: widget.activeColor ??
            sliderThemeData.activeTrackColor ??
            themeData.primaryColor,
        thumbColor: widget.activeColor ??
            sliderThemeData.thumbColor ??
            themeData.primaryColor,
        activeTickColor: sliderThemeData.activeTickColor,
        inactiveTickColor: sliderThemeData.inactiveTickColor,
        disabledActiveTickColor: sliderThemeData.disabledActiveTickColor,
        disabledInactiveTickColor: sliderThemeData.disabledInactiveTickColor,
        activeMinorTickColor: sliderThemeData.activeMinorTickColor,
        inactiveMinorTickColor: sliderThemeData.inactiveMinorTickColor,
        disabledActiveMinorTickColor:
            sliderThemeData.disabledActiveMinorTickColor,
        disabledInactiveMinorTickColor:
            sliderThemeData.disabledInactiveMinorTickColor,
        overlayColor: widget.activeColor?.withOpacity(0.12) ??
            sliderThemeData.overlayColor ??
            themeData.primaryColor.withOpacity(0.12),
        inactiveDivisorColor: widget.activeColor ?? sliderThemeData.inactiveDivisorColor ?? themeData.colorScheme.primary.withOpacity(0.54),
        activeDivisorColor: widget.inactiveColor ?? sliderThemeData.activeDivisorColor ?? themeData.colorScheme.onPrimary.withOpacity(0.54),
        disabledInactiveDivisorColor: sliderThemeData.disabledInactiveDivisorColor ?? themeData.colorScheme.onSurface.withOpacity(0.12),
        disabledActiveDivisorColor: sliderThemeData.disabledActiveDivisorColor ?? themeData.colorScheme.onPrimary.withOpacity(0.12),
        disabledActiveTrackColor: sliderThemeData.disabledActiveTrackColor ?? themeData.colorScheme.onSurface.withOpacity(0.32),
        disabledInactiveTrackColor: sliderThemeData.disabledInactiveTrackColor ?? themeData.colorScheme.onSurface.withOpacity(0.12),
        disabledThumbColor: sliderThemeData.disabledThumbColor,
        activeRegionColor: sliderThemeData.activeRegionColor,
        inactiveRegionColor: widget.inactiveColor ?? sliderThemeData.inactiveRegionColor,
        tooltipBackgroundColor: sliderThemeData.tooltipBackgroundColor,
        trackCornerRadius: sliderThemeData.trackCornerRadius,
        overlayRadius: sliderThemeData.overlayRadius,
        thumbRadius: sliderThemeData.thumbRadius);

    return _RangeSliderRenderObjectWidget(
      key: widget.key,
      min: widget.min,
      max: widget.max,
      values: widget.values,
      onChanged: isActive ? _onChanged : null,
      interval: widget.interval,
      stepSize: widget.stepSize,
      minorTicksPerInterval: widget.minorTicksPerInterval ?? 0,
      showTicks: widget.showTicks,
      showLabels: widget.showLabels,
      showDivisors: widget.showDivisors,
      showTooltip: widget.showTooltip,
      inactiveColor:
          widget.inactiveColor ?? themeData.primaryColor.withOpacity(0.24),
      activeColor: widget.activeColor ?? themeData.primaryColor,
      labelPlacement: widget.labelPlacement ?? LabelPlacement.onTicks,
      numberFormat: widget.numberFormat ?? NumberFormat.decimalPattern(),
      dateIntervalType: widget.dateIntervalType,
      dateFormat: widget.dateFormat,
      labelFormatterCallback:
          widget.labelFormatterCallback ?? _getFormattedLabelText,
      tooltipTextFormatterCallback:
          widget.tooltipTextFormatterCallback ?? _getFormattedTooltipText,
      trackShape: widget.trackShape ?? _SfTrackShape(),
      divisorShape: widget.divisorShape ?? _SfDivisorShape(),
      overlayShape: widget.overlayShape ?? _SfOverlayShape(),
      thumbShape: widget.thumbShape ?? _SfThumbShape(),
      tickShape: widget.tickShape ?? _SfTickShape(),
      minorTickShape: widget.minorTickShape ?? _SfMinorTickShape(),
      sliderThemeData: sliderThemeData,
      state: this,
    );
  }
}

class _RangeSliderRenderObjectWidget extends LeafRenderObjectWidget {
  const _RangeSliderRenderObjectWidget(
      {Key key,
      this.min,
      this.max,
      this.values,
      this.onChanged,
      this.interval,
      this.stepSize,
      this.minorTicksPerInterval,
      this.showTicks,
      this.showLabels,
      this.showDivisors,
      this.showTooltip,
      this.inactiveColor,
      this.activeColor,
      this.labelPlacement,
      this.numberFormat,
      this.dateFormat,
      this.dateIntervalType,
      this.labelFormatterCallback,
      this.tooltipTextFormatterCallback,
      this.trackShape,
      this.divisorShape,
      this.overlayShape,
      this.thumbShape,
      this.tickShape,
      this.minorTickShape,
      this.sliderThemeData,
      this.state})
      : super(key: key);

  final dynamic min;
  final dynamic max;
  final SfRangeValues values;
  final ValueChanged<SfRangeValues> onChanged;
  final double interval;
  final double stepSize;
  final int minorTicksPerInterval;

  final bool showTicks;
  final bool showLabels;
  final bool showDivisors;
  final bool showTooltip;

  final Color inactiveColor;
  final Color activeColor;

  final LabelPlacement labelPlacement;
  final NumberFormat numberFormat;
  final DateIntervalType dateIntervalType;
  final DateFormat dateFormat;
  final SfRangeSliderThemeData sliderThemeData;
  final LabelFormatterCallback labelFormatterCallback;
  final TooltipTextFormatterCallback tooltipTextFormatterCallback;
  final SfDivisorShape divisorShape;
  final SfTrackShape trackShape;
  final SfTickShape tickShape;
  final SfMinorTickShape minorTickShape;
  final SfOverlayShape overlayShape;
  final SfThumbShape thumbShape;
  final _SfRangeSliderState state;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return _RenderRangeSlider(
      min: min,
      max: max,
      values: values,
      onChanged: onChanged,
      minorTicksPerInterval: minorTicksPerInterval,
      interval: interval,
      stepSize: stepSize,
      showTicks: showTicks,
      showLabels: showLabels,
      showDivisors: showDivisors,
      showTooltip: showTooltip,
      inactiveColor: inactiveColor,
      activeColor: activeColor,
      labelPlacement: labelPlacement,
      numberFormat: numberFormat,
      dateFormat: dateFormat,
      dateIntervalType: dateIntervalType,
      labelFormatterCallback: labelFormatterCallback,
      tooltipTextFormatterCallback: tooltipTextFormatterCallback,
      trackShape: trackShape,
      divisorShape: divisorShape,
      overlayShape: overlayShape,
      thumbShape: thumbShape,
      tickShape: tickShape,
      minorTickShape: minorTickShape,
      sliderThemeData: sliderThemeData,
      textDirection: Directionality.of(context),
      mediaQueryData: MediaQuery.of(context),
      state: state,
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, _RenderRangeSlider renderObject) {
    renderObject
      ..min = min
      ..max = max
      ..values = values
      ..onChanged = onChanged
      ..interval = interval
      ..stepSize = stepSize
      ..minorTicksPerInterval = minorTicksPerInterval
      ..showTicks = showTicks
      ..showLabels = showLabels
      ..showDivisors = showDivisors
      ..showTooltip = showTooltip
      ..inactiveColor = inactiveColor
      ..activeColor = activeColor
      ..labelPlacement = labelPlacement
      ..numberFormat = numberFormat
      ..dateFormat = dateFormat
      ..dateIntervalType = dateIntervalType
      ..labelFormatterCallback = labelFormatterCallback
      ..tooltipTextFormatterCallback = tooltipTextFormatterCallback
      ..trackShape = trackShape
      ..divisorShape = divisorShape
      ..overlayShape = overlayShape
      ..thumbShape = thumbShape
      ..tickShape = tickShape
      ..minorTickShape = minorTickShape
      ..sliderThemeData = sliderThemeData
      ..textDirection = Directionality.of(context)
      ..mediaQueryData = MediaQuery.of(context);
  }
}

class _RenderRangeSlider extends RenderProxyBox
    with RelayoutWhenSystemFontsChangeMixin, _SliderMixin {
  _RenderRangeSlider({
    dynamic min,
    dynamic max,
    SfRangeValues values,
    ValueChanged<SfRangeValues> onChanged,
    double interval,
    double stepSize,
    int minorTicksPerInterval,
    bool showTicks,
    bool showLabels,
    bool showDivisors,
    bool showTooltip,
    Color inactiveColor,
    Color activeColor,
    LabelPlacement labelPlacement,
    NumberFormat numberFormat,
    DateFormat dateFormat,
    DateIntervalType dateIntervalType,
    LabelFormatterCallback labelFormatterCallback,
    TooltipTextFormatterCallback tooltipTextFormatterCallback,
    SfTrackShape trackShape,
    SfDivisorShape divisorShape,
    SfOverlayShape overlayShape,
    SfThumbShape thumbShape,
    SfTickShape tickShape,
    SfMinorTickShape minorTickShape,
    SfRangeSliderThemeData sliderThemeData,
    TextDirection textDirection,
    MediaQueryData mediaQueryData,
    @required _SfRangeSliderState state,
  })  : _min = min,
        _max = max,
        _values = values,
        _onChanged = onChanged,
        _interval = interval,
        _stepSize = stepSize,
        _minorTicksPerInterval = minorTicksPerInterval,
        _showTicks = showTicks,
        _showLabels = showLabels,
        _showDivisors = showDivisors,
        _showTooltip = showTooltip,
        _inactiveColor = inactiveColor,
        _activeColor = activeColor,
        _labelPlacement = labelPlacement,
        _numberFormat = numberFormat,
        _dateIntervalType = dateIntervalType,
        _dateFormat = dateFormat,
        _labelFormatterCallback = labelFormatterCallback,
        _tooltipTextFormatterCallback = tooltipTextFormatterCallback,
        _trackShape = trackShape,
        _divisorShape = divisorShape,
        _overlayShape = overlayShape,
        _thumbShape = thumbShape,
        _tickShape = tickShape,
        _minorTickShape = minorTickShape,
        _sliderThemeData = sliderThemeData,
        _textDirection = textDirection,
        _mediaQueryData = mediaQueryData,
        _state = state {
    final GestureArenaTeam team = GestureArenaTeam();

    _dragGestureRecognizer = HorizontalDragGestureRecognizer()
      ..team = team
      ..onStart = _onDragStart
      ..onUpdate = _onDragUpdate
      ..onEnd = _onDragEnd
      ..onCancel = _onDragCancel;

    _tapGestureRecognizer = TapGestureRecognizer()
      ..team = team
      ..onTapDown = _onTapDown
      ..onTapUp = _onTapUp;

    _overlayStartAnimation = CurvedAnimation(
        parent: _state.overlayStartController, curve: Curves.fastOutSlowIn);

    _overlayEndAnimation = CurvedAnimation(
        parent: _state.overlayEndController, curve: Curves.fastOutSlowIn);

    _stateAnimation = CurvedAnimation(
        parent: _state.stateController, curve: Curves.easeInOut);

    _trackOffset = _getTrackOffset();

    if (isDateTime) {
      _minInms = min.millisecondsSinceEpoch.toDouble();
      _maxInms = max.millisecondsSinceEpoch.toDouble();
      _valuesInms = SfRangeValues(
          values.start.millisecondsSinceEpoch.toDouble(),
          values.end.millisecondsSinceEpoch.toDouble());
    }

    _visibleLabels = <String>[];
    _majorTickPositions = <double>[];
    _minorTickPositions = <double>[];
    _updateTextPainter();
  }

  final _SfRangeSliderState _state;

  final TextPainter _textPainter = TextPainter();

  final double _minTrackWidth = kMinInteractiveDimension * 3;

  Offset _trackOffset;

  double _labelsCount;

  double _minInms;

  double _maxInms;

  double _actualHeight;

  SfThumb _activeThumb;

  SfRangeValues _valuesInms;

  bool _showThumbStroke = false;

  bool _willDrawTooltip = false;

  bool _isInteractionEnd = true;

  List<String> _visibleLabels;

  List<double> _majorTickPositions;

  List<double> _minorTickPositions;

  Animation<double> _overlayStartAnimation;

  Animation<double> _overlayEndAnimation;

  Animation<double> _stateAnimation;

  HorizontalDragGestureRecognizer _dragGestureRecognizer;

  TapGestureRecognizer _tapGestureRecognizer;

  bool get isInteractive => onChanged != null;

  bool get isDiscrete => stepSize != null && stepSize > 0;

  Rect get trackRect => _trackShape.getPreferredRect(
      this, _sliderThemeData, Offset.zero, isInteractive);

  Size get actualDivisorSize =>
      _divisorShape.getPreferredSize(_sliderThemeData, isInteractive);

  Size get actualTickSize =>
      _tickShape.getPreferredSize(_sliderThemeData, isInteractive);

  Size get actualMinorTickSize =>
      _minorTickShape.getPreferredSize(_sliderThemeData, isInteractive);

  Size get actualLabelSize => Size.fromHeight(math.max(
      _sliderThemeData.inactiveLabelStyle.fontSize,
      _sliderThemeData.activeLabelStyle.fontSize));

  Size get actualThumbSize =>
      _thumbShape.getPreferredSize(_sliderThemeData, isInteractive);

  Size get actualOverlaySize =>
      _overlayShape.getPreferredSize(_sliderThemeData, isInteractive);

  bool get isDateTime =>
      _min.runtimeType == DateTime &&
      _max.runtimeType == DateTime &&
      _values.start.runtimeType == DateTime &&
      _values.end.runtimeType == DateTime;

  double get _actualTickHeight => _showTicks
      ? _sliderThemeData.tickSize.height +
          (_sliderThemeData.tickOffset != null
              ? _sliderThemeData.tickOffset.dy
              : 0)
      : 0;

  double get _actualMinorTickHeight =>
      _minorTicksPerInterval != null ? actualMinorTickSize.height : 0;

  double get _actualLabelHeight => _showLabels
      ? actualLabelSize.height * _textPainter.textScaleFactor +
          (_sliderThemeData.labelOffset != null
              ? _sliderThemeData.labelOffset.dy
              : 0)
      : 0;

  double get minThumbGap =>
      (actualMax - actualMin) * (8 / trackRect.width).clamp(0.0, 1.0);

  double get actualMin => isDateTime ? _minInms : _min;

  double get actualMax => isDateTime ? _maxInms : _max;

  SfRangeValues get actualValues => isDateTime ? _valuesInms : _values;

  dynamic get min => _min;
  dynamic _min;
  set min(dynamic value) {
    if (_min == value) {
      return;
    }
    _min = value;

    if (_min is DateTime) {
      _minInms = _min.millisecondsSinceEpoch.toDouble();
    }
    markNeedsPaint();
  }

  dynamic get max => _max;
  dynamic _max;
  set max(dynamic value) {
    if (_max == value) {
      return;
    }
    _max = value;

    if (_max is DateTime) {
      _maxInms = _max.millisecondsSinceEpoch.toDouble();
    }
    markNeedsPaint();
  }

  SfRangeValues get values => _values;
  SfRangeValues _values;
  set values(SfRangeValues values) {
    if (isDiscrete) {
      if (_activeThumb == SfThumb.start) {
        final dynamic startThumb = _getActualValue(value: values.start);
        values = SfRangeValues(startThumb, values.end);
      } else {
        final dynamic endThumb = _getActualValue(value: values.end);
        values = SfRangeValues(values.start, endThumb);
      }
    }

    if (_values == values) {
      return;
    }

    _values = values;
    if (isDateTime) {
      _valuesInms = SfRangeValues(
          _values.start.millisecondsSinceEpoch.toDouble(),
          _values.end.millisecondsSinceEpoch.toDouble());
    }
    markNeedsPaint();
  }

  ValueChanged<SfRangeValues> get onChanged => _onChanged;
  ValueChanged<SfRangeValues> _onChanged;
  set onChanged(ValueChanged<SfRangeValues> value) {
    if (value == _onChanged) {
      return;
    }
    final bool wasInteractive = isInteractive;
    _onChanged = value;
    if (wasInteractive != isInteractive) {
      if (isInteractive) {
        _state.stateController.forward();
      } else {
        _state.stateController.reverse();
      }
      markNeedsPaint();
    }
  }

  double get interval => _interval;
  double _interval;
  set interval(double value) {
    if (_interval == value) {
      return;
    }
    _interval = value;
    markNeedsPaint();
  }

  double get stepSize => _stepSize;
  double _stepSize;
  set stepSize(double value) {
    if (_stepSize == value) {
      return;
    }
    _stepSize = value;
    markNeedsPaint();
  }

  int get minorTicksPerInterval => _minorTicksPerInterval;
  int _minorTicksPerInterval;
  set minorTicksPerInterval(int value) {
    if (_minorTicksPerInterval == value) {
      return;
    }
    _minorTicksPerInterval = value;
    markNeedsPaint();
  }

  bool get showTicks => _showTicks;
  bool _showTicks;
  set showTicks(bool value) {
    if (_showTicks == value) {
      return;
    }
    _showTicks = value;
    markNeedsLayout();
  }

  bool get showLabels => _showLabels;
  bool _showLabels;
  set showLabels(bool value) {
    if (_showLabels == value) {
      return;
    }
    _showLabels = value;
    markNeedsLayout();
  }

  bool get showDivisors => _showDivisors;
  bool _showDivisors;
  set showDivisors(bool value) {
    if (_showDivisors == value) {
      return;
    }
    _showDivisors = value;
    markNeedsLayout();
  }

  bool get showTooltip => _showTooltip;
  bool _showTooltip;
  set showTooltip(bool value) {
    if (_showTooltip == value) {
      return;
    }
    _showTooltip = value;
  }

  Color get inactiveColor => _inactiveColor;
  Color _inactiveColor;
  set inactiveColor(Color value) {
    if (_inactiveColor == value) {
      return;
    }
    _inactiveColor = value;
    markNeedsPaint();
  }

  Color get activeColor => _activeColor;
  Color _activeColor;
  set activeColor(Color value) {
    if (_activeColor == value) {
      return;
    }
    _activeColor = value;
    markNeedsPaint();
  }

  LabelPlacement get labelPlacement => _labelPlacement;
  LabelPlacement _labelPlacement;
  set labelPlacement(LabelPlacement value) {
    if (_labelPlacement == value) {
      return;
    }
    _labelPlacement = value;
    markNeedsPaint();
  }

  NumberFormat get numberFormat => _numberFormat;
  NumberFormat _numberFormat;
  set numberFormat(NumberFormat value) {
    if (_numberFormat == value) {
      return;
    }
    _numberFormat = value;
    markNeedsPaint();
  }

  DateIntervalType get dateIntervalType => _dateIntervalType;
  DateIntervalType _dateIntervalType;
  set dateIntervalType(DateIntervalType value) {
    if (_dateIntervalType == value) {
      return;
    }
    _dateIntervalType = value;
    markNeedsPaint();
  }

  DateFormat get dateFormat => _dateFormat;
  DateFormat _dateFormat;
  set dateFormat(DateFormat value) {
    if (_dateFormat == value) {
      return;
    }
    _dateFormat = value;
    markNeedsPaint();
  }

  LabelFormatterCallback get labelFormatterCallback => _labelFormatterCallback;
  LabelFormatterCallback _labelFormatterCallback;
  set labelFormatterCallback(LabelFormatterCallback value) {
    if (_labelFormatterCallback == value) {
      return;
    }
    _labelFormatterCallback = value;
    markNeedsSemanticsUpdate();
  }

  TooltipTextFormatterCallback get tooltipTextFormatterCallback =>
      _tooltipTextFormatterCallback;
  TooltipTextFormatterCallback _tooltipTextFormatterCallback;
  set tooltipTextFormatterCallback(TooltipTextFormatterCallback value) {
    if (_tooltipTextFormatterCallback == value) {
      return;
    }
    _tooltipTextFormatterCallback = value;
    markNeedsSemanticsUpdate();
  }

  SfDivisorShape get divisorShape => _divisorShape;
  SfDivisorShape _divisorShape;
  set divisorShape(SfDivisorShape value) {
    if (_divisorShape == value) {
      return;
    }
    _divisorShape = value;
    markNeedsLayout();
  }

  SfTrackShape get trackShape => _trackShape;
  SfTrackShape _trackShape;
  set trackShape(SfTrackShape value) {
    if (_trackShape == value) {
      return;
    }
    _trackShape = value;
    markNeedsLayout();
  }

  SfTickShape get tickShape => _tickShape;
  SfTickShape _tickShape;
  set tickShape(SfTickShape value) {
    if (_tickShape == value) {
      return;
    }
    _tickShape = value;
    markNeedsLayout();
  }

  SfMinorTickShape get minorTickShape => _minorTickShape;
  SfMinorTickShape _minorTickShape;
  set minorTickShape(SfMinorTickShape value) {
    if (_minorTickShape == value) {
      return;
    }
    _minorTickShape = value;
    markNeedsLayout();
  }

  SfOverlayShape get overlayShape => _overlayShape;
  SfOverlayShape _overlayShape;
  set overlayShape(SfOverlayShape value) {
    if (_overlayShape == value) {
      return;
    }
    _overlayShape = value;
    markNeedsLayout();
  }

  SfThumbShape get thumbShape => _thumbShape;
  SfThumbShape _thumbShape;
  set thumbShape(SfThumbShape value) {
    if (_thumbShape == value) {
      return;
    }
    _thumbShape = value;
    markNeedsLayout();
  }

  SfRangeSliderThemeData get sliderThemeData => _sliderThemeData;
  SfRangeSliderThemeData _sliderThemeData;
  set sliderThemeData(SfRangeSliderThemeData value) {
    if (_sliderThemeData == value) {
      return;
    }
    _sliderThemeData = value;
    markNeedsPaint();
  }

  TextDirection get textDirection => _textDirection;
  TextDirection _textDirection;
  set textDirection(TextDirection value) {
    if (_textDirection == value) {
      return;
    }
    _textDirection = value;
    _updateTextPainter();
    markNeedsPaint();
  }

  MediaQueryData get mediaQueryData => _mediaQueryData;
  MediaQueryData _mediaQueryData;
  set mediaQueryData(MediaQueryData value) {
    if (_mediaQueryData == value) {
      return;
    }
    _mediaQueryData = value;
    _updateTextPainter();
    markNeedsLayout();
  }

  Offset _getTrackOffset() {
    final double dx = <double>[
          actualOverlaySize.width,
          actualThumbSize.width,
          actualTickSize.width,
          actualMinorTickSize.width
        ].reduce(math.max) /
        2;

    final double dy = <double>[
          actualOverlaySize.height,
          actualThumbSize.height,
          actualDivisorSize.height,
          _sliderThemeData.trackHeight
        ].reduce(math.max) /
        2;

    return Offset(dx, dy);
  }

  void _onTapDown(TapDownDetails details) {
    _beginInteraction(globalToLocal(details.globalPosition).dx);
  }

  void _onTapUp(TapUpDetails details) {
    _endInteraction();
  }

  void _onDragStart(DragStartDetails details) {
    _beginInteraction(globalToLocal(details.globalPosition).dx);
  }

  void _onDragUpdate(DragUpdateDetails details) {
    _isInteractionEnd = false;
    _updateRangeValues(globalToLocal(details.globalPosition).dx);
    markNeedsPaint();
  }

  void _onDragEnd(DragEndDetails details) {
    _endInteraction();
  }

  void _onDragCancel() {
    _endInteraction();
  }

  void _endInteraction() {
    if (!_isInteractionEnd) {
      _state.overlayStartController.reverse();
      _state.overlayEndController.reverse();
      _willDrawTooltip = false;
      _isInteractionEnd = true;
      markNeedsPaint();
    }
  }

  void _beginInteraction(double xPosition) {
    final double startPosition =
        _valueToFactor(actualValues.start) * trackRect.width + trackRect.left;
    final double endPosition =
        _valueToFactor(actualValues.end) * trackRect.width + trackRect.left;
    final double leftThumbWidth = (startPosition - xPosition).abs();
    final double rightThumbWidth = (endPosition - xPosition).abs();

    _willDrawTooltip = true;
    if (rightThumbWidth > leftThumbWidth) {
      _activeThumb = SfThumb.start;
      _state.overlayStartController.forward();
    } else {
      _activeThumb = SfThumb.end;
      _state.overlayEndController.forward();
    }
    _isInteractionEnd = false;
    _updateRangeValues(xPosition);
    markNeedsPaint();
  }

  void _updateTextPainter() {
    _textPainter
      ..textDirection = _textDirection
      ..textScaleFactor = _mediaQueryData.textScaleFactor;
  }

  dynamic _getValueFromPosition(double position) {
    double valueFactor =
        ((position - trackRect.left) / trackRect.width).clamp(0.0, 1.0);
    if (_textDirection == TextDirection.rtl) {
      valueFactor = 1.0 - valueFactor;
    }

    final dynamic actualValue = _factorToValue(valueFactor);
    if (isDateTime) {
      return _getActualValue(valueInms: actualValue);
    } else {
      return _getActualValue(value: actualValue);
    }
  }

  double _valueToPosition(double value) {
    return _valueToFactor(value) * trackRect.width + trackRect.left;
  }

  double _positionToFactor(double position) {
    final double factor =
        ((position - trackRect.left) / trackRect.width).clamp(0.0, 1.0);
    return (_textDirection == TextDirection.rtl) ? 1.0 - factor : factor;
  }

  double _valueToFactor(dynamic value) {
    final double factor =
        (value == null) ? 0.0 : ((value - actualMin) / (actualMax - actualMin));
    return (_textDirection == TextDirection.rtl) ? 1.0 - factor : factor;
  }

  double _factorToValue(double factor) {
    return factor * (actualMax - actualMin) + actualMin;
  }

  void _updateRangeValues(double xPosition) {
    double start;
    double end;
    SfRangeValues newValues;
    final double factor = _positionToFactor(xPosition);
    if (isDateTime) {
      start = values.start.millisecondsSinceEpoch.toDouble();
      end = values.end.millisecondsSinceEpoch.toDouble();
    } else {
      start = values.start;
      end = values.end;
    }

    final double value = lerpDouble(actualMin, actualMax, factor);
    if (_activeThumb == SfThumb.start) {
      newValues = values._copyWith(start: math.min(value, end - minThumbGap));
    } else {
      newValues = values._copyWith(end: math.max(value, start + minThumbGap));
    }
    _showThumbStroke =
        (_valueToFactor(start) - _valueToFactor(end)).abs() * trackRect.width <
            actualThumbSize.height;

    onChanged(newValues);
  }

  dynamic _getActualValue({dynamic value, double valueInms}) {
    if (isDiscrete) {
      if (!isDateTime) {
        final double maxMinDiff = _max - _min;
        value = _factorToValue(
            ((_valueToFactor(value) * (maxMinDiff / stepSize)).round() /
                    (maxMinDiff / stepSize))
                .clamp(0.0, 1.0));
      }
    }

    return isDateTime
        ? (value ?? DateTime.fromMillisecondsSinceEpoch(valueInms.toInt()))
        : value;
  }

  void _generateVisibleLabels() {
    _visibleLabels.clear();
    _majorTickPositions.clear();
    _minorTickPositions.clear();

    String label;
    dynamic currentValue = _min;
    _labelsCount = (isDateTime
                ? _getDateTimeDifference(_min, _max, _dateIntervalType)
                : _max - _min)
            .toDouble() /
        _interval;
    for (int i = 0; i <= _labelsCount; i++) {
      label = _labelFormatterCallback(
          currentValue, _getFormattedText(currentValue));
      _visibleLabels.add(label);

      final double labelPosition = _valueToFactor(
              isDateTime ? currentValue.millisecondsSinceEpoch : currentValue) *
          (trackRect.width);
      if (!_majorTickPositions.contains(labelPosition)) {
        _majorTickPositions.add(labelPosition);
      }

      currentValue = isDateTime
          ? _getNextDate(currentValue, _dateIntervalType, _interval)
          : currentValue + _interval;
    }
    final int majorTicksCount = _majorTickPositions.length;
    if (_minorTicksPerInterval > 0) {
      if (isDateTime) {
        double minorTickPosition;
        DateTime nextDate = _getNextDate(_min, _dateIntervalType, _interval);
        DateTime currentActualDate =
            _getNextDate(nextDate, _dateIntervalType, -_interval);
        for (int i = 1; i <= majorTicksCount; i++) {
          // Need to divide the region based on _minorTicksPerInterval. So, added 1 with _minorTicksPerInterval.
          final double intervalDiff = _getDateTimeDifference(
                  currentActualDate, nextDate, _dateIntervalType) /
              (_minorTicksPerInterval + 1);
          // On dividing region equally between two dates, then it will be equal to 1.0.
          if (intervalDiff == 1.0) {
            // To get nextDate of minorTick, need to pass intervals. So, iteration value is used as interval and is started with 1.
            for (double j = 1; j <= _minorTicksPerInterval; j++) {
              final DateTime nextMinorDate =
                  _getNextDate(currentActualDate, _dateIntervalType, j);
              minorTickPosition = _valueToPosition(
                      nextMinorDate.millisecondsSinceEpoch.toDouble()) -
                  trackRect.left;
              _minorTickPositions.add(minorTickPosition);
            }
          } else {
            final double minorPositionDiff = (nextDate.millisecondsSinceEpoch -
                    currentActualDate.millisecondsSinceEpoch) /
                (_minorTicksPerInterval + 1);
            for (int j = 1; j <= _minorTicksPerInterval; j++) {
              minorTickPosition = _valueToPosition(
                      currentActualDate.millisecondsSinceEpoch +
                          (j * minorPositionDiff)) -
                  trackRect.left;
              _minorTickPositions.add(minorTickPosition);
            }
          }
          currentActualDate = nextDate;
          nextDate =
              _getNextDate(currentActualDate, _dateIntervalType, interval);
        }
      } else {
        for (int i = 0; i <= majorTicksCount - 1; i++) {
          final double minorPositionDiff = ((i + 1 < majorTicksCount
                      ? _majorTickPositions[i + 1]
                      : trackRect.width) -
                  _majorTickPositions[i]) /
              (_minorTicksPerInterval + 1);
          for (int j = 1; j <= _minorTicksPerInterval; j++) {
            _minorTickPositions
                .add(_majorTickPositions[i] + j * minorPositionDiff);
          }
        }
      }
    }
  }

  @override
  void systemFontsDidChange() {
    super.systemFontsDidChange();
    _textPainter.markNeedsLayout();
    _updateTextPainter();
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    _overlayStartAnimation.addListener(markNeedsPaint);
    _overlayEndAnimation.addListener(markNeedsPaint);
    _stateAnimation.addListener(markNeedsPaint);
  }

  @override
  void detach() {
    _overlayStartAnimation.removeListener(markNeedsPaint);
    _overlayEndAnimation.removeListener(markNeedsPaint);
    _stateAnimation.removeListener(markNeedsPaint);

    super.detach();
  }

  @override
  bool hitTestSelf(Offset position) => isInteractive;

  @override
  bool hitTestChildren(BoxHitTestResult result, {Offset position}) => false;

  @override
  void setupParentData(RenderObject child) {
    if (child.parentData is! BoxParentData) {
      child.parentData = BoxParentData();
    }
  }

  @override
  void performLayout() {
    _actualHeight = math.max(
        2 * _trackOffset.dy,
        _trackOffset.dy +
            _sliderThemeData.trackHeight / 2 +
            math.max(_actualTickHeight, _actualMinorTickHeight) +
            _actualLabelHeight);

    size = Size(
      constraints.hasBoundedWidth
          ? constraints.maxWidth
          : _minTrackWidth + 2 * _trackOffset.dx,
      constraints.hasBoundedHeight ? constraints.maxHeight : _actualHeight,
    );
    if (interval != null && interval > 0) {
      _generateVisibleLabels();
    }
  }

  @override
  void handleEvent(PointerEvent event, HitTestEntry entry) {
    // Checked [_isInteractionEnd] for avoiding multi touch.
    if (_isInteractionEnd && event.down && event is PointerDownEvent) {
      _dragGestureRecognizer.addPointer(event);
      _tapGestureRecognizer.addPointer(event);
    }
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    final Offset actualTrackOffset = Offset(
        offset.dx,
        offset.dy +
            (size.height - _actualHeight) / 2 +
            _trackOffset.dy -
            _sliderThemeData.trackHeight / 2);

    // Drawing track.
    final Rect _trackRect = _trackShape.getPreferredRect(
        this, _sliderThemeData, actualTrackOffset, isInteractive);
    final dynamic thumbStartPosition =
        _valueToFactor(actualValues.start) * _trackRect.width;
    final dynamic thumbEndPosition =
        _valueToFactor(actualValues.end) * _trackRect.width;

    final Offset startThumbCenter =
        Offset(_trackRect.left + thumbStartPosition, _trackRect.center.dy);
    final Offset endThumbCenter =
        Offset(_trackRect.left + thumbEndPosition, _trackRect.center.dy);

    _trackShape.paint(
        context, actualTrackOffset, startThumbCenter, endThumbCenter,
        isEnabled: isInteractive,
        parentBox: this,
        themeData: _sliderThemeData,
        animation: _stateAnimation,
        textDirection: _textDirection);

    if (_interval != null && _interval > 0) {
      int _dateTimePos = 0;
      final double dx = _trackRect.left;
      final double dy = _trackRect.top;
      final double halfTrackHeight = _trackRect.height / 2;
      final double divisorRadius = _divisorShape
              .getPreferredSize(_sliderThemeData, isInteractive)
              .width /
          2;
      final double tickRadius =
          _tickShape.getPreferredSize(_sliderThemeData, isInteractive).width /
              2;

      double textvalue = isDateTime ? 0.0 : _min;
      int minorTickIndex = 0;
      final double maxRange = isDateTime ? _labelsCount : _max;

      while (textvalue <= maxRange) {
        final double tickPosition = _majorTickPositions[_dateTimePos];

        // Drawing divisors.
        if (showDivisors) {
          Offset divisorCenter =
              Offset(dx + tickPosition, dy + halfTrackHeight);
          if (_majorTickPositions[_dateTimePos] == 0.0)
            divisorCenter =
                Offset(dx + tickPosition + divisorRadius, dy + halfTrackHeight);
          else if (_majorTickPositions[_dateTimePos] == _trackRect.width)
            divisorCenter =
                Offset(dx + tickPosition - divisorRadius, dy + halfTrackHeight);
          _divisorShape.paint(
              context, divisorCenter, startThumbCenter, endThumbCenter,
              isEnabled: isInteractive,
              parentBox: this,
              themeData: _sliderThemeData,
              animation: _stateAnimation,
              textDirection: _textDirection);
        }

        // Drawing ticks.
        if (showTicks) {
          Offset actualTickOffset =
              Offset(dx + tickPosition, dy + _trackRect.height) +
                  (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          if (_majorTickPositions[_dateTimePos] == 0.0)
            actualTickOffset =
                Offset(dx + tickPosition + tickRadius, dy + _trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          else if (_majorTickPositions[_dateTimePos] == _trackRect.width)
            actualTickOffset =
                Offset(dx + tickPosition - tickRadius, dy + _trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));

          _tickShape.paint(
              context, actualTickOffset, startThumbCenter, endThumbCenter,
              isEnabled: isInteractive,
              parentBox: this,
              themeData: _sliderThemeData,
              animation: _stateAnimation,
              textDirection: _textDirection);
        }

        if (_minorTicksPerInterval > 0) {
          for (int j = 0; j < _minorTicksPerInterval; j++) {
            final double currentMinorTickPosition =
                _minorTickPositions[minorTickIndex];
            minorTickIndex++;

            if (currentMinorTickPosition < trackRect.width &&
                currentMinorTickPosition > 0) {
              final Offset actualTickOffset = Offset(
                      dx + currentMinorTickPosition, dy + _trackRect.height) +
                  (_sliderThemeData.tickOffset ?? const Offset(0, 0));
              _minorTickShape.paint(
                  context, actualTickOffset, startThumbCenter, endThumbCenter,
                  isEnabled: isInteractive,
                  parentBox: this,
                  themeData: _sliderThemeData,
                  animation: _stateAnimation,
                  textDirection: _textDirection);
            }
          }
        }

        // Drawing label.
        if (showLabels) {
          final dynamic labelText = _visibleLabels[_dateTimePos];
          final bool isRTL = _textDirection == TextDirection.rtl;
          double offsetX = dx + tickPosition;
          if (labelPlacement == LabelPlacement.betweenTicks) {
            offsetX += ((_dateTimePos + 1 <= _labelsCount
                        ? _majorTickPositions[_dateTimePos + 1]
                        : (isRTL ? _trackRect.left : _trackRect.width)) -
                    tickPosition) /
                2;
            if (isRTL
                ? offsetX <= _trackRect.left
                : offsetX - dx >= _trackRect.width) {
              break;
            }
          }
          final Offset actualLabelOffset = Offset(
                offsetX,
                dy + _trackRect.height + _actualTickHeight,
              ) +
              (_sliderThemeData.labelOffset ?? const Offset(0, 0));

          _drawText(context, actualLabelOffset, startThumbCenter, labelText,
              endThumbCenter: endThumbCenter,
              parentBox: this,
              themeData: _sliderThemeData,
              animation: _stateAnimation,
              textPainter: _textPainter,
              textDirection: _textDirection);
        }

        final dynamic intervalDiff = isDateTime ? 1 : interval;
        textvalue += intervalDiff;
        _dateTimePos += 1;
      }
    }

    final Paint paint = Paint()..isAntiAlias = true;
    final bool isLeftThumbActive = _activeThumb == SfThumb.start;
    Offset thumbCenter = isLeftThumbActive ? endThumbCenter : startThumbCenter;
    // Drawing thumb.
    _thumbShape.paint(context, thumbCenter,
        isEnabled: isInteractive,
        parentBox: this,
        themeData: _sliderThemeData,
        animation: _stateAnimation,
        textDirection: _textDirection,
        thumb: isLeftThumbActive ? SfThumb.end : SfThumb.start);

    thumbCenter = isLeftThumbActive ? startThumbCenter : endThumbCenter;
    // Drawing overlay.
    _overlayShape.paint(context, thumbCenter,
        isEnabled: isInteractive,
        parentBox: this,
        themeData: _sliderThemeData,
        animation:
            isLeftThumbActive ? _overlayStartAnimation : _overlayEndAnimation,
        thumb: _activeThumb);

    // Drawing thumb.
    _thumbShape.paint(context, thumbCenter,
        isEnabled: isInteractive,
        parentBox: this,
        themeData: _sliderThemeData,
        animation: _stateAnimation,
        textDirection: _textDirection,
        thumb: _activeThumb);

    if (_thumbShape is _SfThumbShape && _showThumbStroke) {
      paint
        ..color = Colors.white
        ..style = PaintingStyle.stroke
        ..strokeWidth = 0.5;

      if (isLeftThumbActive) {
        context.canvas
            .drawCircle(startThumbCenter, sliderThemeData.thumbRadius, paint);
      } else {
        context.canvas
            .drawCircle(endThumbCenter, sliderThemeData.thumbRadius, paint);
      }
    }

    if (_showTooltip && _willDrawTooltip) {
      paint
        ..color = _sliderThemeData.tooltipBackgroundColor
        ..style = PaintingStyle.fill
        ..strokeWidth = 0;
      _drawTooltip(
          context,
          offset,
          Offset(actualTrackOffset.dx + _trackOffset.dx, actualTrackOffset.dy),
          startThumbCenter,
          endThumbCenter,
          paint,
          _trackRect);
    }
  }

  String _getFormattedText(dynamic value) {
    if (isDateTime) {
      return _dateFormat != null ? _dateFormat.format(value) : value.toString();
    }
    return _numberFormat.format(value);
  }

  // Here 10 is a gap between tooltip nose and thumb.
  double get tooltipStartY => actualThumbSize.height / 2 + 10;

  void _drawTooltip(
      PaintingContext context,
      Offset offset,
      Offset trackStartOffset,
      Offset startThumbCenter,
      Offset endThumbCenter,
      Paint paint,
      Rect trackRect) {
    Offset thumbCenter =
        _activeThumb == SfThumb.start ? endThumbCenter : startThumbCenter;
    dynamic actualText = _getValueFromPosition(thumbCenter.dx - offset.dx);
    String tooltipText =
        tooltipTextFormatterCallback(actualText, _getFormattedText(actualText));
    TextSpan textSpan =
        TextSpan(text: tooltipText, style: _sliderThemeData.tooltipTextStyle);
    _textPainter.text = textSpan;
    _textPainter.layout();
    Path path = getTooltipPath(_textPainter.size + _tooltipTextPadding,
        tooltipStartY, trackStartOffset, trackRect, thumbCenter.dx, size);
    context.canvas
        .drawPath(path, paint..color = _sliderThemeData.tooltipBackgroundColor);

    final double leftPadding = _tooltipTextPadding.dx / 2;
    final double minLeftdx = trackStartOffset.dx + leftPadding;
    double dx = _getActualTooltipTextX(
        minLeftdx, thumbCenter, _textPainter.width / 2, leftPadding, trackRect);
    final double dy = tooltipStartY +
        _tooltipTriangleHeight +
        (path.getBounds().size.height - _tooltipTriangleHeight) / 2 +
        _textPainter.height / 2 +
        trackRect.height / 2;
    _textPainter.paint(context.canvas, Offset(dx, thumbCenter.dy - dy));

    thumbCenter =
        _activeThumb == SfThumb.start ? startThumbCenter : endThumbCenter;
    actualText = _getValueFromPosition(thumbCenter.dx - offset.dx);
    tooltipText =
        tooltipTextFormatterCallback(actualText, _getFormattedText(actualText));
    textSpan =
        TextSpan(text: tooltipText, style: _sliderThemeData.tooltipTextStyle);
    _textPainter.text = textSpan;
    _textPainter.layout();
    path = getTooltipPath(_textPainter.size + _tooltipTextPadding,
        tooltipStartY, trackStartOffset, trackRect, thumbCenter.dx, size);
    context.canvas.drawPath(path, paint);

    dx = _getActualTooltipTextX(
        minLeftdx, thumbCenter, _textPainter.width / 2, leftPadding, trackRect);
    _textPainter.paint(context.canvas, Offset(dx, thumbCenter.dy - dy));
  }

  double _getActualTooltipTextX(double minLeftdx, Offset thumbCenter,
      double halfTextWidth, double leftPadding, Rect trackRect) {
    final double dx = thumbCenter.dx - halfTextWidth;
    return dx >= minLeftdx
        ? thumbCenter.dx + halfTextWidth > trackRect.right
            ? trackRect.right - _textPainter.width - leftPadding
            : dx
        : minLeftdx;
  }
}
