part of sliders;

/// Base class for [SfRangeSlider] and [SfRangeSelector] track shapes.
abstract class SfTrackShape {
  const SfTrackShape();

  /// Returns the size based on the values passed to it.
  Rect getPreferredRect(RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData, Offset offset, bool isEnabled) {
    final double maxRadius = math.max(themeData.overlayRadius,
        math.max(themeData.thumbRadius, themeData.tickSize.width / 2));
    final double left = offset.dx + maxRadius;
    final double top = offset.dy;
    final double width = parentBox.size.width - 2 * maxRadius;
    final double height = themeData.trackHeight;

    return Rect.fromLTWH(left, top, width, height);
  }

  /// Paints the track based on the values passed to it.
  void paint(
    PaintingContext context,
    Offset offset,
    Offset startThumbCenter,
    Offset endThumbCenter, {
    bool isEnabled,
    RenderProxyBox parentBox,
    SfRangeSliderThemeData themeData,
    Animation<double> animation,
    TextDirection textDirection,
  }) {
    final ColorTween activeTrackColorTween = ColorTween(
        begin: themeData.disabledActiveTrackColor,
        end: themeData.activeTrackColor);
    final ColorTween inactiveTrackColorTween = ColorTween(
        begin: themeData.disabledInactiveTrackColor,
        end: themeData.inactiveTrackColor);

    final Radius radius = Radius.circular(themeData.trackCornerRadius);
    final Rect actualTrackRect =
        getPreferredRect(parentBox, themeData, offset, isEnabled);

    final Paint paint = Paint()
      ..isAntiAlias = true
      ..strokeWidth = 0
      ..color = inactiveTrackColorTween.evaluate(animation);

    // Drawing inactive track.
    Rect trackRect = Rect.fromLTRB(actualTrackRect.left, actualTrackRect.top,
        startThumbCenter.dx, actualTrackRect.bottom);
    final RRect leftRRect = RRect.fromRectAndCorners(trackRect,
        topLeft: radius, bottomLeft: radius);
    context.canvas.drawRRect(leftRRect, paint);

    // Drawing active track.
    paint.color = activeTrackColorTween.evaluate(animation);
    trackRect = Rect.fromLTRB(startThumbCenter.dx, actualTrackRect.top,
        endThumbCenter.dx, actualTrackRect.bottom);
    final RRect centerRRect = RRect.fromRectAndCorners(trackRect,
        topLeft: Radius.zero,
        topRight: Radius.zero,
        bottomLeft: Radius.zero,
        bottomRight: Radius.zero);
    context.canvas.drawRRect(centerRRect, paint);

    // Drawing inactive track.
    paint.color = inactiveTrackColorTween.evaluate(animation);
    trackRect = Rect.fromLTRB(endThumbCenter.dx, actualTrackRect.top,
        actualTrackRect.width + actualTrackRect.left, actualTrackRect.bottom);
    final RRect rightRRect = RRect.fromRectAndCorners(trackRect,
        topLeft: Radius.zero,
        topRight: radius,
        bottomLeft: Radius.zero,
        bottomRight: radius);
    context.canvas.drawRRect(rightRRect, paint);
  }
}

/// Base class for [SfRangeSlider] and [SfRangeSelector] thumb shapes.
abstract class SfThumbShape {
  const SfThumbShape();

  /// Returns the size based on the values passed to it.
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return Size.fromRadius(themeData.thumbRadius);
  }

  /// Paints the thumb based on the values passed to it.
  void paint(PaintingContext context, Offset center,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection,
      SfThumb thumb}) {
    context.canvas.drawCircle(
        center,
        getPreferredSize(themeData, isEnabled).width / 2,
        Paint()
          ..isAntiAlias = true
          ..strokeWidth = 0
          ..color =
              isEnabled ? themeData.thumbColor : themeData.disabledThumbColor);
  }
}

/// Base class for [SfRangeSlider] and [SfRangeSelector] divisors shapes.
abstract class SfDivisorShape {
  const SfDivisorShape();

  /// Returns the size based on the values passed to it.
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return Size.fromRadius(themeData.trackHeight / 2);
  }

  /// Paints the divisors based on the values passed to it.
  void paint(PaintingContext context, Offset center, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    bool isInactive;
    switch (textDirection) {
      case TextDirection.ltr:
        isInactive =
            center.dx < startThumbCenter.dx || center.dx > endThumbCenter.dx;
        break;
      case TextDirection.rtl:
        isInactive =
            center.dx > startThumbCenter.dx || center.dx < endThumbCenter.dx;
        break;
    }
    final Color begin = isInactive
        ? themeData.disabledInactiveDivisorColor
        : themeData.disabledActiveDivisorColor;
    final Color end = isInactive
        ? themeData.inactiveDivisorColor
        : themeData.activeDivisorColor;
    final Paint paint = Paint()
      ..color = ColorTween(begin: begin, end: end).evaluate(animation);
    final double divisorRadius =
        getPreferredSize(themeData, isEnabled).width / 2;
    context.canvas.drawCircle(center, divisorRadius, paint);
  }
}

/// Base class for [SfRangeSlider] and [SfRangeSelector] overlay shapes.
abstract class SfOverlayShape {
  const SfOverlayShape();

  /// Returns the size based on the values passed to it.
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return Size.fromRadius(themeData.overlayRadius);
  }

  /// Paints the overlay based on the values passed to it.
  void paint(PaintingContext context, Offset center,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      SfThumb thumb}) {
    final double radius = getPreferredSize(themeData, isEnabled).width / 2;
    final Tween<double> tween = Tween<double>(begin: 0.0, end: radius);

    context.canvas.drawCircle(
        center,
        tween.evaluate(animation),
        Paint()
          ..isAntiAlias = true
          ..strokeWidth = 0
          ..color = themeData.overlayColor);
  }
}

/// Base class for [SfRangeSlider] and [SfRangeSelector] major tick shapes.
abstract class SfTickShape {
  const SfTickShape();

  /// Returns the size based on the values passed to it.
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return Size.copy(themeData.tickSize);
  }

  /// Paints the major ticks based on the values passed to it.
  void paint(PaintingContext context, Offset offset, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    bool isInactive;
    final Size tickSize = getPreferredSize(themeData, isEnabled);
    switch (textDirection) {
      case TextDirection.ltr:
        isInactive =
            offset.dx < startThumbCenter.dx || offset.dx > endThumbCenter.dx;
        break;
      case TextDirection.rtl:
        isInactive =
            offset.dx > startThumbCenter.dx || offset.dx < endThumbCenter.dx;
        break;
    }
    final Color begin = isInactive
        ? themeData.disabledInactiveTickColor
        : themeData.disabledActiveTickColor;
    final Color end =
        isInactive ? themeData.inactiveTickColor : themeData.activeTickColor;
    final Paint paint = Paint()
      ..isAntiAlias = true
      ..strokeWidth = tickSize.width
      ..color = ColorTween(begin: begin, end: end).evaluate(animation);
    context.canvas.drawLine(
        offset, Offset(offset.dx, offset.dy + tickSize.height), paint);
  }
}

/// Base class for [SfRangeSlider] and [SfRangeSelector] minor tick shapes.
abstract class SfMinorTickShape {
  const SfMinorTickShape();

  /// Returns the size based on the values passed to it.
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return Size.copy(themeData.minorTickSize);
  }

  /// Paints the minor ticks based on the values passed to it.
  void paint(PaintingContext context, Offset offset, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    bool isInactive;
    final Size minorTickSize = getPreferredSize(themeData, isEnabled);
    switch (textDirection) {
      case TextDirection.ltr:
        isInactive =
            offset.dx < startThumbCenter.dx || offset.dx > endThumbCenter.dx;
        break;
      case TextDirection.rtl:
        isInactive =
            offset.dx > startThumbCenter.dx || offset.dx < endThumbCenter.dx;
        break;
    }
    final Color begin = isInactive
        ? themeData.disabledInactiveMinorTickColor
        : themeData.disabledActiveMinorTickColor;
    final Color end = isInactive
        ? themeData.inactiveMinorTickColor
        : themeData.activeMinorTickColor;
    final Paint paint = Paint()
      ..isAntiAlias = true
      ..strokeWidth = minorTickSize.width
      ..color = ColorTween(begin: begin, end: end).evaluate(animation);
    context.canvas.drawLine(
        offset, Offset(offset.dx, offset.dy + minorTickSize.height), paint);
  }
}

class _SfTrackShape extends SfTrackShape {
  @override
  Rect getPreferredRect(RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData, Offset offset, bool isEnabled) {
    return super.getPreferredRect(parentBox, themeData, offset, isEnabled);
  }

  @override
  void paint(PaintingContext context, Offset offset, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    super.paint(context, offset, startThumbCenter, endThumbCenter,
        isEnabled: isEnabled,
        parentBox: parentBox,
        themeData: themeData,
        animation: animation,
        textDirection: textDirection);
  }
}

class _SfThumbShape extends SfThumbShape {
  @override
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return super.getPreferredSize(themeData, isEnabled);
  }

  @override
  void paint(PaintingContext context, Offset center,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection,
      SfThumb thumb}) {
    super.paint(context, center,
        isEnabled: isEnabled,
        parentBox: parentBox,
        themeData: themeData,
        animation: animation,
        textDirection: textDirection);
  }
}

class _SfDivisorShape extends SfDivisorShape {
  @override
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return super.getPreferredSize(themeData, isEnabled);
  }

  @override
  void paint(PaintingContext context, Offset center, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    super.paint(context, center, startThumbCenter, endThumbCenter,
        isEnabled: isEnabled,
        parentBox: parentBox,
        themeData: themeData,
        animation: animation,
        textDirection: textDirection);
  }
}

class _SfOverlayShape extends SfOverlayShape {
  @override
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return super.getPreferredSize(themeData, isEnabled);
  }

  @override
  void paint(PaintingContext context, Offset center,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      bool isDiscrete,
      List<dynamic> values,
      TextDirection textDirection,
      SfThumb thumb}) {
    super.paint(context, center,
        isEnabled: isEnabled,
        parentBox: parentBox,
        themeData: themeData,
        animation: animation);
  }
}

class _SfTickShape extends SfTickShape {
  @override
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return super.getPreferredSize(themeData, isEnabled);
  }

  @override
  void paint(PaintingContext context, Offset offset, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    super.paint(context, offset, startThumbCenter, endThumbCenter,
        isEnabled: isEnabled,
        parentBox: parentBox,
        themeData: themeData,
        animation: animation,
        textDirection: textDirection);
  }
}

class _SfMinorTickShape extends SfMinorTickShape {
  @override
  Size getPreferredSize(SfRangeSliderThemeData themeData, bool isEnabled) {
    return super.getPreferredSize(themeData, isEnabled);
  }

  @override
  void paint(PaintingContext context, Offset offset, Offset startThumbCenter,
      Offset endThumbCenter,
      {bool isEnabled,
      RenderProxyBox parentBox,
      SfRangeSliderThemeData themeData,
      Animation<double> animation,
      TextDirection textDirection}) {
    super.paint(context, offset, startThumbCenter, endThumbCenter,
        isEnabled: isEnabled,
        parentBox: parentBox,
        themeData: themeData,
        animation: animation,
        textDirection: textDirection);
  }
}
